#include <LibSdl2/include/SDL.h>
#include "lillibsdl.h"

typedef struct Sdl2Context_s {
	SDL_Window *win;
	SDL_Renderer *rend;
} Sdl2Context;

int InitSdl2(int w, int h, Sdl2Context *pContext) {
	pContext->win = SDL_CreateWindow("Image Viewer", SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 640, 480, SDL_WINDOW_SHOWN);
    if (!pContext->win)
		return -1;

	pContext->rend = SDL_CreateRenderer(pContext->win, -1, SDL_RENDERER_ACCELERATED);
	if (!pContext->rend)
		return -1;

	return 0;
}

void ExitSdl2(Sdl2Context *pContext) {
	SDL_DestroyRenderer(pContext->rend);
	SDL_DestroyWindow(pContext->win);
}

#if (defined(_WIN32) || defined(_WIN64) || defined(WIN32) || defined(WIN64) || defined(_WIN32_WCE))
#include <Windows.h>
int WINAPI WinMain(HINSTANCE hinst, HINSTANCE hinst2, LPSTR cmd, int num) {
#else
int main(int argc, char** argv) {
#endif

	SDL_Event e;
	int running = 1;
	Sdl2Context context;
	SDL_Texture *pTexture = 0;

	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) == -1)
		return 1;

	if (InitSdl2(640, 480, &context) == -1) {
		SDL_Quit();
		return 1;
	}

	while (running) {
		while (SDL_PollEvent(&e)) {
			if (e.type == SDL_QUIT)
				running = 0;
			if (e.type == SDL_KEYDOWN && e.key.keysym.sym == SDLK_ESCAPE)
				running = 0;
			if (e.type == SDL_DROPFILE && e.drop.file) {
				if (LoadSdlTexFromFile(&pTexture, context.rend, e.drop.file, 0, 0) != LE_NONE)
					pTexture = 0;
			}
		}
		
		SDL_SetRenderDrawBlendMode(context.rend, SDL_BLENDMODE_NONE);

		SDL_RenderClear(context.rend);
		if (pTexture)
			SDL_RenderCopy(context.rend, pTexture, 0, 0);

		SDL_RenderPresent(context.rend);
	}

	ExitSdl2(&context);
	SDL_Quit();
	return 0;
}
