#include "LilCodecJpeg.h"
#include "LilUtilFormat.h"
#include "LilConvert.h"
#include "LilColorspace.h"
#include "jpge.h"

namespace lil {

namespace {




//-----------------------------------------------------------------------------------------------------------
// NOTE: Modified "Loeffler IDCT modified algorithm"

/* #define 1108     sqrt(2) * cos(6 * Pi / 16)
 * #define 2676     sqrt(2) * sin(6 * Pi / 16)
 * #define 1703     cos(3 * Pi / 16)
 * #define 1138     sin(3 * Pi / 16)
 * #define 2009     cos(Pi / 16)
 * #define 400      sin(Pi / 16)
 */
 
/* Function prototypes used to implement Loeffler IDCT modified algorithm */
#define mult(n, x) (((n) * (x) + (1 << 10)) >> 11)

LIL_INLINE void IDCT_1D (lil_i32 i0, lil_i32 i1, lil_i32 i2, lil_i32 i3, lil_i32 i4, lil_i32 i5, lil_i32 i6, lil_i32 i7,
								lil_i32 &o0, lil_i32 &o1, lil_i32 &o2, lil_i32 &o3, lil_i32 &o4, lil_i32 &o5, lil_i32 &o6, lil_i32 &o7)
{
	// NOTE shift is 11 bits...

    lil_i32 s0, s1, s2, s3, s4, s5, s6, s7;
    lil_i32 tmp = 0;
 
    /* Stage 1 */
	s1 = i0 - i4;
	s0 = i0 + i4;
    /* rot(sqrt(2), 6, Y[k + 2, Y[k + 6, &s2, &s3); */
    tmp = mult(1108, i2 + i6);
 
    s2 = tmp - mult(3784, i6);
    s3 = tmp + mult(1568, i2);
	s4 = i1 - i7;
	s7 = i1 + i7;
      
    /* Multiply by sqrt(2) */
    s5 = mult(2896, i3);
    s6 = mult(2896, i5);
     
    /* Stage 2 */
	o3 = s0 - s3;
	o0 = s0 + s3;
	o2 = s1 - s2;
	o1 = s1 + s2;
	lil_i32 t6 = s4 - s6;
	lil_i32 t4 = s4 + s6;
	lil_i32 t5 = s7 - s5;
	lil_i32 t7 = s7 + s5;
 
    /* Stage 3 */
    /* rot(1, 3, s2[4, s2[7, &s3[4, &s3[7); */
    tmp = mult(1703, t4 + t7);
 
    o4 = tmp - mult(2841, t7);
    o7 = tmp - mult(565, t4);
 
    /* rot(1, 1, s2[5, s2[6, &s3[5, &s3[6); */
    tmp = mult(2009, t5 + t6);
 
    o5 = tmp - mult(2409, t6);
    o6 = tmp - mult(1609, t5);
 
    // Stage 4 - External! Use o* outputs
}

// Tried column->row - no benefit.
void IDCT (lil_i32 *input, lil_u8 *output, lil_u32 stride)
{
	for (lil_u8 k = 0; k < 64; k += 8) {
        // IDCT 1D ROW
		lil_i32 *i = input + k;
		// This happens enough to be beneficial
		if ((i[1] | i[2] | i[3] | i[4] | i[5] | i[6] | i[7]) == 0)
			i[1] = i[2] = i[3] = i[4] = i[5] = i[6] = i[7] = i[0];
		else {
		    lil_i32 t0, t1, t2, t3, t4, t5, t6, t7;
			IDCT_1D(i[0], i[1], i[2], i[3], i[4], i[5], i[6], i[7],
				t0, t1, t2, t3, t4, t5, t6, t7);
			i[7] = t0 - t7;
			i[0] = t0 + t7;
			i[6] = t1 - t6;
			i[1] = t1 + t6;
			i[5] = t2 - t5;
			i[2] = t2 + t5;
			i[4] = t3 - t4;
			i[3] = t3 + t4;
		}
	}
    for (lil_u8 n = 0; n < 8; n++) {
        // IDCT 1D COLUMN
		lil_i32 *i = input + n;
		IDCT_1D(i[0], i[8], i[16], i[24], i[32], i[40], i[48], i[56],
			i[0], i[8], i[16], i[24], i[32], i[40], i[48], i[56]);
		lil_u8 *pOut = output + n;

#define PROCESS(x) ((((x) + 4) >> 3) + 128)
#define CLAMP(x) (t ? (~t >> 24) : x)
#define ASSIGN(dest, x) r = PROCESS(x); t = r >> 8; *dest = static_cast<lil_u8>(CLAMP(r)); dest += stride;
		lil_u32 t;
		lil_i32 r;
		ASSIGN(pOut, i[0] + i[56]);
		ASSIGN(pOut, i[8] + i[48]);
		ASSIGN(pOut, i[16] + i[40]);
		ASSIGN(pOut, i[24] + i[32]);
		ASSIGN(pOut, i[24] - i[32]);
		ASSIGN(pOut, i[16] - i[40]);
		ASSIGN(pOut, i[8] - i[48]);
		ASSIGN(pOut, i[0] - i[56]);
#undef ASSIGN
#undef CLAMP
#undef PROCESS
    }
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
// MSB-LSB BIT READER WITH JPEG MODIFICATIONS
//-----------------------------------------------------------------------------------------------------------
class JpegBitReadMsb {
public:
	LIL_INLINE JpegBitReadMsb(const LilIoDesc *io) : io(io), hasBits(0), currBits(0) { }

	//-----------------------------------------------------------------------------------------------------------
	LIL_INLINE void AlignToByte() {
		// Keep the lowest multiple of 8 bits, since they are the most recent whole bytes
		hasBits -= hasBits & 7;  // effectively goes to the next byte in logic
		currBits &= (1 << hasBits) - 1;
	}

	//-----------------------------------------------------------------------------------------------------------
	LIL_INLINE void ReadBits(lil_u32 &output, lil_u8 numBits) {
		while (hasBits < numBits) {
			currBits = (currBits << 8) | Read8(io);
			hasBits += 8;
		}
		hasBits -= numBits;
		output = (currBits >> hasBits);
		//currBits &= ((1 << hasBits) - 1);
		currBits ^= output << hasBits;
	}

	// [Spec F.2.2.5 - Next Bit Read]
	// [Spec B.1.1.5 - Entropy Coded Data Segments]
	LIL_INLINE void ReadEntropyBits(lil_u32 &output, lil_u8 numBits) {
		while (hasBits < numBits) {
			lil_u8 b = Read8(io);
			if (b == 0xff) {
				lil_u8 c = Read8(io);
				if (c) {
					// [Spec B.1.1.5] '1' bits are used to pad to integer byte count before reset marker.
					// So discard those bits if we found a reset marker...
					hasBits = 0;
					currBits = 0;  // output assignment depends on this not having random bits!!!
					output = 0xff00 | c;
					return;
				}
			}
			currBits = (currBits << 8) | b;
			hasBits += 8;
		}
		hasBits -= numBits;
		output = currBits >> hasBits;
		//currBits &= ((1 << hasBits) - 1);
		currBits ^= output << hasBits;
	}

	LIL_INLINE lil_u8 GetNumBits() const { return hasBits; }

	//-----------------------------------------------------------------------------------------------------------
	// Note that input is considered to be LSB-MSB aligned, even though this reader is
	// for MSB-LSB aligned input.
	LIL_INLINE void UndoBits(lil_u32 input, lil_u8 numBits) {
		currBits |= (input & ((1 << numBits) - 1)) << hasBits;
		hasBits += numBits;
	}

private:
	const LilIoDesc *io;
	lil_u8 hasBits;
	lil_u32 currBits;
};



//-----------------------------------------------------------------------------------------------------------
// HUFFMAN
//-----------------------------------------------------------------------------------------------------------
#define HUFFMAN_LUT_BITS 9
#define HUFFMAN_LUT_SIZE (1 << HUFFMAN_LUT_BITS)
struct HuffmanMsb {
	// If the "code" is a single byte, it's implied that there can only be up to 256 codes.
	// Use lil_u16 data type because we will ensure codes be no more than 16 mBits long - though
	// with uncontrolled huffman they could be up to n - 1 mBits long.
	lil_u16 code[256];
	lil_u32 maxCode[24];  // For bit counts 0 - 16
	lil_u8 startCode[17];  // For bit coutn 0 - 16
	// The code lengths (number of mBits) for each of the above codes (up to 256 codes).
	// The last code + 1 always has a value of 0 to denote that THAT'S IT.
	// See below for generating the code table to see why.
	lil_u8 size[257];
	// If there's only 256 codes, then there's of course only 256 values.  These are in the
	// same order that the codes are in, so when you match a code, use that index to index
	// this table.
	lil_u8 val[256];
	
	// @@ Maybe this isn't the best way... but it'll be a good start.
	// Problems may be 
	// - Cache thrashing
	// - Stack not large enough especially on non-modern major cpus
	lil_u8 fastUndo[HUFFMAN_LUT_SIZE];
	lil_u8 fastDict[HUFFMAN_LUT_SIZE];

	//Never used...
	//LIL_INLINE void Undo(JpegBitReadMsb &reader, lil_u32 bits, lil_u8 numBits) {
	//	reader.UndoBits(bits, numBits);
	//}

	LIL_INLINE lil_u32 JpegDecode(JpegBitReadMsb &reader) {
		lil_u32 total = 0;
		lil_u8 numBits = reader.GetNumBits();
		if (!numBits)
			numBits = 8;
		reader.ReadEntropyBits(total, numBits);
		while (total >= maxCode[numBits]) {
			lil_u32 bits;
			reader.ReadEntropyBits(bits, 8);
			total = (total << 8) | bits;
			numBits += 8;
		}

		// See if we can do a fast lookup
		if (numBits <= HUFFMAN_LUT_BITS) {
			lil_u8 shift = (HUFFMAN_LUT_BITS - numBits);
			lil_u32 msbBits = total << shift;
			reader.UndoBits(total, fastUndo[msbBits] - shift);
			return fastDict[msbBits];
		}
		lil_u8 shift = numBits - HUFFMAN_LUT_BITS;
		lil_u32 msbBits = total >> shift;
		if (msbBits < maxCode[HUFFMAN_LUT_BITS]) {
			reader.UndoBits(total, fastUndo[msbBits] + shift);
			return fastDict[msbBits];
		}

		// Slow lookup isn't THAT slow... we can start where we left off above
		lil_u8 count = HUFFMAN_LUT_BITS;
		numBits -= HUFFMAN_LUT_BITS;
		lil_u32 bits = total;
		do {
			numBits--;
			total = bits >> numBits;
			count++;
		} while (total >= maxCode[count]);
		reader.UndoBits(bits, numBits);

		// Now we have the huffman code... we need to translate it to a value
		for (lil_u32 i = startCode[count]; i < 256; i++)
			if (code[i] == total)
				return val[i];
		return 0xffffffff;  // no good!!!
	}

	// See in spec
	LIL_INLINE lil_u32 JpegReceiveExtend(JpegBitReadMsb &reader, lil_u32 numBits) {
		LIL_ASSERT(numBits <= 16);
		lil_u32 huff = 0;
		reader.ReadEntropyBits(huff, static_cast<lil_u8>(numBits));
		// extend
		lil_u32 test = 1 << (numBits - 1);
		if (huff < test)  // sign extend
			huff += (-1 << numBits) + 1;
		return huff;
	}

	// Requires "val" to have been filled already (from jpeg file)
	LilErrorType GenCanonJpegDecode(const lil_u8 *lens) {
		// Generate size table from lengths array
		// (Jpeg spec makes it confusing... it's pretty obvious to get it from the "lens" array,
		// which contains 16 values of number of codes with that length (1 through 16 mBits) )
		lil_u8 k = 0;
		for (lil_u8 i = 0; i < 16; i++)
			for (lil_u32 j = 0; j < lens[i]; j++)
				size[k++] = i + 1;
		size[k] = 0; // marks end

		// Generate code table according to Jpeg spec
		k = 0;
		lil_u16 currCode = 0;
		memset(maxCode, 0xff, sizeof(maxCode));
		maxCode[0] = 0;
		startCode[0] = 0;

		// Need a valid code for each maxCode

		memset(code, 0, sizeof(code));
		for (lil_u32 si = 1; si <= 16; si++) {
			startCode[si] = k;
			while (size[k] == si)
				code[k++] = currCode++;
			if (currCode > (1 << si))  // Make sure code is in bit range
				return LE_BADIMAGEDATA;
			maxCode[si] = currCode;
			// This is basically the part that makes it a canonical jpeg huffman code...
			currCode <<= 1;
		}

		// Now code and size are ordered according to increasing huffman code numeric value and length.

		// Generate test dictionary
		// Only code[0] can possibly have value 0. Since val[0] may NOT be 0, we
		// don't want to re-assign dict[0] later!
		lil_u8 dict[HUFFMAN_LUT_SIZE];
		if (code[0] < HUFFMAN_LUT_SIZE)
			dict[code[0]] = val[0];
		for (lil_u32 i = 1; i < 256; i++)
			if (code[i] && code[i] < HUFFMAN_LUT_SIZE)
				dict[code[i]] = val[i];

		for (lil_u32 j = 0; j < HUFFMAN_LUT_SIZE; j++) {
			lil_u8 bits = 1;
			while (bits <= HUFFMAN_LUT_BITS && (j >> (HUFFMAN_LUT_BITS - bits)) >= maxCode[bits])
				bits++;
			fastDict[j] = dict[j >> (HUFFMAN_LUT_BITS - bits)];
			fastUndo[j] = (HUFFMAN_LUT_BITS - bits);
		}

		return LE_NONE;
	}

/*	bool GenCanonJpegEncode() {
		for (lil_u32 k = 0; size[k]; k++) {
			lil_u8 i = val[k];

		// ?? See page 52 & 53 in jpeg standard
		return true;
	}*/
};



//-----------------------------------------------------------------------------------------------------------
// JPEG IO
//-----------------------------------------------------------------------------------------------------------
LilErrorType LilIoJpegRead(lil_io_handle handle, size_t byteCount, void *outMemPtr) {
	LilCodecJpeg *pHandle = reinterpret_cast<LilCodecJpeg *>(handle);
	size_t offset = pHandle->mTransferOffset;
	lil_i8 *pOut8 = reinterpret_cast<lil_i8 *>(outMemPtr);
	lil_i8 *pIn8 = pHandle->mStrip.GetPtr<lil_i8>();
	size_t size = pHandle->mStrip.GetSize();
	lil_u32 x = pHandle->mTransferX;
	lil_u32 w = pHandle->mImage->data.width * pHandle->mImage->format.numChannels;
	LilErrorType err = LE_NONE;
	lil_u32 bpl = LilImageGetBpl(pHandle->mImage);

	// @@ Won't work on 12 bit images yet
	LIL_ASSERT(pHandle->mImage->format.bitsPerChannel == 8);

	while (byteCount) {
		// Get next set of data
		if (!pHandle->IsReady()) {
			err = pHandle->DecodeImageData();
			if (err != LE_NONE)
				return err;
			offset = 0;
			pHandle->mReady = true;
		}

		// A single scan line won't have more than 4 billion pixels
		lil_u32 count = static_cast<lil_u32>(byteCount < w - x ? byteCount : w - x);
		memcpy(pOut8, &pIn8[offset], count);
		pOut8 += count;
		x += count;
		offset += count;
		byteCount -= count;
		if (x == w) {
			byteCount -= bpl - w;
			pOut8 += bpl - w;

			x = 0;
			offset += pHandle->mStripBpl - w;
			if (offset == size)
				pHandle->mReady = false;
		}
	}

	// @@ I'm not sure if it's valid for jpegs to be bigger than 4 gb compressed?
	pHandle->mTransferOffset = static_cast<lil_u32>(offset);
	pHandle->mTransferX = x;

	return LE_NONE;
}

//} // namespace



//-----------------------------------------------------------------------------------------------------------
// JPEG
//-----------------------------------------------------------------------------------------------------------
LilCodecJpeg::LilCodecJpeg(const LilIoDesc *pIoInput, LilImageDesc *pImage, const LilFormatDesc *pFormat) :
	mIo(pIoInput), mImage(pImage), pFinalFormat(pFormat), mReader(0),
	mHeight(0), mWidth(0), mBits(0), mCh(0), mThumbW(0), mThumbH(0), 
	mFrameMarker(MARKER_NONE_PADDING), mMcuW(0), mMcuH(0), mMcuCountX(0), mMcuCountY(0), mJfifVersion(0),
	mRestartInterval(0), mRestartCurr(0), mScanNumComp(0),
	mReady(false), mDecoding(false), mStripBpl(0), mTransferOffset(0), mTransferX(0),
	mIter(-1), mMcuX(0), mMcuY(0)
{
	ioOutput.handle = reinterpret_cast<lil_io_handle>(this);
	ioOutput.readFunc = LilIoJpegRead;
	ioOutput.writeFunc = 0;
	ioOutput.seekFunc = 0;
	ioOutput.tellFunc = 0;

	for (lil_u32 i = 0; i < 4; i++) {
		mHuffAc[i] = 0;
		mHuffDc[i] = 0;
	}

	mReader = new JpegBitReadMsb(pIoInput);

	for (lil_u32 i = 0; i < 4; i++) {
		mHuffAc[i] = new HuffmanMsb;
		mHuffDc[i] = new HuffmanMsb;
	}
}

//-----------------------------------------------------------------------------------------------------------
LilCodecJpeg::~LilCodecJpeg() {
	delete mReader;
	for (lil_u32 i = 0; i < 4; i++) {
		delete mHuffAc[i];
		delete mHuffDc[i];
	}
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilCodecJpeg::DecodeStart() {
	if (Read16be(mIo) != MARKER_START_OF_IMAGE)
		return LE_BADHEADER;

	// Spec says SOI marker resets mStartInterval specifically
	mRestartInterval = 0;
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilCodecJpeg::DecodeImageInfo() {
	LilErrorType err = LE_NONE;
	lil_u16 headerLen = 0;
	lil_u8 code = 0;

	bool done = false;
	while (!done) {
		lil_u16 marker = ReadMarker();
		// [Spec 4.2 - Lossy and lossless overview]
		// [Spec 4.5 - Modes of operation overview]
		// [Spec 4.6 - Entropy coding methods overview]
		// [Spec 4.11 - Characteristics of different coding processes]
		switch (marker) {
		// [Spec B.2.2 - Frame header]
		// [Spec 4.3 - DCT overview]
		// non-differential, huffman coding
		case MARKER_BASELINE_DCT:
		// non-differential, huffman coding
		case MARKER_EXT_SEQ_DCT_H:
			{
				ReadFbe(mIo, "21221", &headerLen, &mBits, &mHeight, &mWidth, &mCh);
				if (headerLen != 8 + 3 * mCh || mWidth == 0)
					return LE_BADHEADER;				
				//if (mCh == 0 || (marker == MARKER_PROG_DCT_A && mCh > 4))
				if (mCh == 0 || mCh > 4)  // Regardless of spec... we wouldn't know what to do with extra components anyway...
					return LE_BADHEADER;
				// [Spec 4.7 - Sample Precision]
				if ((marker == MARKER_BASELINE_DCT && mBits != 8) || 
					((marker == MARKER_EXT_SEQ_DCT_H || marker == MARKER_PROG_DCT_A) && mBits != 8 && mBits != 12) ||
					(marker == MARKER_LOSSLESS_A && mBits < 2 && mBits > 16))
					return LE_BADHEADER;
				lil_u8 maxSampH = 0, maxSampV = 0;
				for (lil_u32 i = 0; i < mCh; i++) {
					ReadFbe(mIo, "111", &mComp[i].id, &code, &mComp[i].quantI);
					mComp[i].sampFactH = code >> 4;
					mComp[i].sampFactV = code & 0x0f;
					maxSampH = GetMax(mComp[i].sampFactH, maxSampH);
					maxSampV = GetMax(mComp[i].sampFactV, maxSampV);
					if (mComp[i].sampFactH == 0 || mComp[i].sampFactH > 4 || mComp[i].sampFactV == 0 || mComp[i].sampFactV > 4)
						return LE_BADHEADER;
					if (mComp[i].quantI > 3 || (marker == MARKER_LOSSLESS_A && mComp[i].quantI > 0))
						return LE_BADHEADER;
				}
				// Figure out number of MCUs wide and high the image will be.
				// [Spec 4.8 - Interleaved multiple components]
				// [Spec A.2 - Interleaved Minimum Coded Unit order]
				// [Spec B.2.2 (see above - info on sampling factors)]
				if (mCh > 1) {
					mMcuW = maxSampH * 8;
					mMcuH = maxSampV * 8;
				}
				else {
					mMcuW = 8;
					mMcuH = 8;
				}
				mMcuCountX = (mWidth + (mMcuW - 1)) / mMcuW;
				mMcuCountY = (mHeight + (mMcuH - 1)) / mMcuH;

				// [Spec A.1.1 - Sampling factors]
				for (lil_u32 i = 0; i < mCh; i++) {
					// It's important we don't chop off the last few pixels in this division!!
					mComp[i].w = (mWidth * mComp[i].sampFactH + maxSampH - 1) / maxSampH;
					mComp[i].h = (mHeight * mComp[i].sampFactV + maxSampV - 1) / maxSampV;
				}
				if (mCh == 3)
					mCh = 4;

				if (!SetImageSize(&mImage->data, mWidth, mHeight, &err))
					return err;
				if (!SetFormat(&mImage->format, LDF_SIMPLE, mBits, mCh, 4, false,
					mCh == 1 ? LCO_SINGLE : LCO_BGRA, false, &err))
					return err;
				if (!SetExtendedInfo(&mImage->extended, 1, LC_JPG_BASELINE, LD_INCHES, 0, 0, &err))
					return err;
				mImage->extended.numPages = 1;
				mImage->extended.pageIndex = 0;
				mImage->extended.compression = LC_JPG_BASELINE;
				mImage->extended.fileFormat.jpg.subsampling[0] = 4;
				mImage->extended.fileFormat.jpg.subsampling[1] = mImage->extended.fileFormat.jpg.subsampling[2] = 0;
				if (mCh > 1) {
					mImage->extended.fileFormat.jpg.subsampling[1] = 4 / (mComp[0].sampFactH / mComp[1].sampFactH);
					if (mComp[2].sampFactV == mComp[0].sampFactV) {
						if (mComp[2].sampFactH == mComp[1].sampFactH)
							mImage->extended.fileFormat.jpg.subsampling[2] = mImage->extended.fileFormat.jpg.subsampling[1];
						else if (mComp[2].sampFactH < mComp[1].sampFactH)
							mImage->extended.fileFormat.jpg.subsampling[2] = 4 / (mComp[0].sampFactH / mComp[2].sampFactH);
					}
					// else it already = 0
				}
				mFrameMarker = marker;

				// Allow for extra bytes from MCUs so we can directly decode into strip memory
				mStripBpl = RoundUpToNearest(LilImageGetBpl(mImage), 8 * maxSampH * mCh);
				if (!mStrip.Realloc(&mImage->memFuncs, maxSampV * 8 * mStripBpl))
					return LE_NOTENOUGHMEMORY;
				done = true;
			}
			break;

		// non-differential, huffman coding
		case MARKER_PROG_DCT_H:
		// [Spec 4.4 - Lossless overview]
		// [Spec A.4 - Point transform for lossless and hierarchical]
		// non-differential, huffman coding
		case MARKER_LOSSLESS_H:
		// NOTE arithmetic coding features appear to have patent issues??
		// non-differential, arithmetic coding
		case MARKER_EXT_SEQ_DCT_A:
		// non-differential, arithmetic coding
		case MARKER_PROG_DCT_A:
		// non-differential, arithmetic coding
		case MARKER_LOSSLESS_A:

			return LE_UNSUPPORTEDFILEFEATURE;
		
		default:
			err = InterpretMarker(marker);
			if (err != LE_NONE)
				return err;
			break;
		}
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
// Interprets frame data now that we have it loaded
LilErrorType LilCodecJpeg::DecodeImageData() {
	LilErrorType err = LE_NONE;
	lil_u16 headerLen = 0;
	lil_u8 code = 0;

	bool done = false;
	while (!done) {
		if (mDecoding) {
			for (lil_u32 i = 0; i < mMcuCountX; i++) {
				err = DecodeMCU();
				if (err != LE_NONE) {
					done = true;
					break;
				}
			}
			if (mMcuY != mMcuCountY)
				done = true;
		}
		else {
			lil_u16 marker = ReadMarker();
			switch (marker) {
			// [Spec 4.10 - One Image, One Frame, Multiple Scans]
			// [Spec A.2 - Source data ordering]
			case MARKER_START_OF_SCAN:
				{
					headerLen = Read16be(mIo);
					mScanNumComp = Read8(mIo);  // count of components in scan
					if (mScanNumComp != 1 && mScanNumComp != 3)
						return LE_BADCOLORSPACE;
					lil_u32 sumHV = 0;
					lil_u8 start = 0;  // Start of spectral for dct (should be 0!) or predictor selection for lossless
					lil_u8 end = 0;  // End of spectral for dct (should be 63!) or meaningless for lossless (make it 0)
					for (lil_u32 i = 0; i < mScanNumComp; i++) {
						lil_u8 compId = Read8(mIo);
						lil_u8 compI = 0;
						for (; compI < mCh; compI++)
							if (mComp[compI].id == compId)
								break;
						if (compI == mCh)
							return LE_BADHEADER;
						sumHV += mComp[compI].sampFactH * mComp[compI].sampFactV;
						mScanCompOrder[i] = compI;
						code = Read8(mIo);
						if ((code & 0x0f) > 3 || (code >> 4) > 3)
							return LE_BADHEADER;

						mComp[compI].huffDcI = code >> 4;
						mComp[compI].huffAcI = code & 0x0f;
					}
					// Jpeg spec restriction on scans with > 1 components in it
					if (mScanNumComp > 1 && sumHV > 10)
						return LE_BADHEADER;
					ReadFbe(mIo, "111", &start, &end, &code);
					if (mFrameMarker == MARKER_PROG_DCT_H || mFrameMarker == MARKER_BASELINE_DCT) {
						if (start != 0)
							return LE_BADHEADER;
						// I guess this is wrong in some files... sometimes it's 0
						//if (end != 63)
						//	return LE_BADHEADER;
						// Spec says always 0 for DCT
						if (code != 0)
							return LE_BADHEADER; 
					}

					DecodeReset();

					mDecoding = true;
				}
				break;
			case MARKER_END_OF_IMAGE:
				done = true;
				break;
			default:
				err = InterpretMarker(marker);
				if (err != LE_NONE)
					return err;
				break;
			}
		}
	}
	return err;
}

//-----------------------------------------------------------------------------------------------------------
lil_u16 LilCodecJpeg::ReadMarker() {
	lil_u32 marker;
	mReader->ReadBits(marker, 16);
	// Spec claims this is possible...
	while (marker == MARKER_NONE_PADDING) {
		lil_u32 val;
		mReader->ReadBits(val, 8);
		marker = (marker & 0xff00) | val;
	}
	return static_cast<lil_u16>(marker);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilCodecJpeg::InterpretMarker(lil_u16 marker) {
	lil_u8 code = 0, zero = 0, units = 0;
	lil_u16 headerLen = 0, xUnits = 0, yUnits = 0;
	lil_u32 id = 0;

	switch (marker) {
	case MARKER_APP0:
		// Some encoders incorrectly save jfif jpegs with wrong marker order, so we just take them
		// as they come....
		ReadFbe(mIo, "241", &headerLen, &id, &zero);
		if (id == ID_JFIF) {
			ReadFbe(mIo, "2122 11", &mJfifVersion, &units, &xUnits, &yUnits, &mThumbW, &mThumbH);
			if (zero != ZERO_8 || headerLen != 16 + mThumbW * mThumbH * 3)
				return LE_BADHEADER;
			// According to recommendatmIons from some sources, the major number is the one to pay
			// attentmIon to and expect incorrect minor numbers occasmIonally I guess??
			if ((mJfifVersion & 0xff00) > (SUPPORTED_VERSION & 0xff00))
				return LE_UNSUPPORTEDVERSION;
			mIo->seekFunc(mIo->handle, mThumbW * mThumbH * 3, LS_RELATIVE);  // @@ Skip thumbnail for now
		}
		else if (id == ID_JFXX) {
			if (mJfifVersion < 0x0102)  // Must be at least 0102 to use ID_JFXX
				return LE_BADHEADER;
			// Check for extensmIons IF it is a valid versmIon to contain them.
			code = Read8(mIo);
			mIo->seekFunc(mIo->handle, headerLen - 8, LS_RELATIVE);  // @@ Skip thumbnail for now
		}
		else
			mIo->seekFunc(mIo->handle, headerLen - 7, LS_RELATIVE);
		break;
	case MARKER_COMMENT:
		headerLen = Read16be(mIo);
		// @@ not impl yet
		mIo->seekFunc(mIo->handle, headerLen - 2, LS_RELATIVE);
		break;
	case MARKER_QUANTIZATION:
		{
			// This can store multiple quantizatmIon tables in 1 marker, or the encoder could
			// have stored multiple tables in just as many markers.
			lil_i32 len = Read16be(mIo) - 2;  // signed so we can detect something wrong with size...
			while (len > 0) {
				code = Read8(mIo);  // high 4 is 8 or 16 bit, low 4 is which table index (0-3)
				// @@@ These are in zigzag scan order.   Should they be unzigzagged yet??
				if (code >> 4 == PRECISION_8_BIT)
					for (lil_u8 i = 0; i < 64; i++) 
						mQt[code & 3][i] = Read8(mIo);
				else if (code >> 4 == PRECISION_16_BIT)
					for (lil_u8 i = 0; i < 64; i++) 
						mQt[code & 3][i] = Read16be(mIo);
				else
					return LE_BADHEADER;
				len -= 65 + (code >> 4);
			}
			if (len != 0)
				return LE_BADHEADER;
		}
		break;
	case MARKER_HUFFMAN_SPEC:
		{
			// This can store multiple huffman tables in 1 marker, or the encoder could
			// have stored multiple tables in just as many markers.
			lil_i32 len = Read16be(mIo) - 2;  // signed so we can detect something wrong with size...
			while (len > 0) {
				code = Read8(mIo);
				lil_u8 lens[16];
				// Get the number of codes of lengths 1 - 16
				ReadN(mIo, 16, lens);
				
				HuffmanMsb *pHuff = 0;
				if ((code >> 4) > 1 || (code & 0x0f) >= 4)
					return LE_BADHEADER;
				if (code >> 4 == 0)
					pHuff = mHuffDc[code & 0x0f];
				else
					pHuff = mHuffAc[code & 0x0f];

				lil_u32 k = 0;
				for (lil_u32 i = 0; i < 16; i++) {
					ReadN(mIo, lens[i], pHuff->val + k);
					k += lens[i];
				}
				if (k > 256)
					return LE_BADHEADER;
				len -= 17 + k;

				LilErrorType err = pHuff->GenCanonJpegDecode(lens);
				if (err != LE_NONE)
					return err;
			}
			if (len != 0)
				return LE_BADHEADER;
		}
		break;
	case MARKER_RESTART_INTERVAL:
		headerLen = Read16be(mIo);
		if (headerLen != 4)
			return LE_BADHEADER;
		mRestartInterval = Read16be(mIo);
		break;
	case MARKER_ARITH_SPEC:
	default:
		// These are Restart Interval Termination markers.
		// They have no other parameters.
		if (marker >= MARKER_RESTART_0 && marker <= MARKER_RESTART_7) {
			mRestartCurr--;
			DecodeReset();
			break;
		}
		// APP# blocks
		if (marker >= MARKER_APP0 && marker <= MARKER_APPF) {
			headerLen = Read16be(mIo);
			mIo->seekFunc(mIo->handle, headerLen - 2, LS_RELATIVE);
			break;
		}

		// Otherwise it's something we don't currently support.
		return LE_UNSUPPORTEDFILEFEATURE;
	}
	return LE_NONE;
}

//-----------------------------------------------------------------------------------------------------------
void LilCodecJpeg::DecodeReset() {
	for (lil_u32 i = 0; i < 4; i++)
		mComp[i].pred = 0;
	mRestartCurr = 0;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LilCodecJpeg::DecodeMCU() {
	// [Spec 4.2 - Lossy and lossless overview]
	// [Spec 4.5 - Modes of operation overview]
	// [Spec 4.6 - Entropy coding methods overview]
	// [Spec 4.3 - DCT overview]
	// [Spec 4.4 Fig 6 - DCT decoder simplified diagram]
	// [Spec 4.8 - Multiple component control/Interleaving components]
	// [Spec 4.8.2 - MCU related to interleaving]
	// [Spec 4.9 - Restart marker]
	// [Spec A.2 - MCU Data Unit interleaving, subsampling, organization]
	mIter++;
	for (lil_u32 i = 0; i < mScanNumComp; i++) {
		// Now, start decode!
		// 1) DECODE BLOCK - IE A "DATA UNIT"
		// 1A) DECODE DC
		lil_i32 table[64];  // This is the data unit
		//lil_u8 tableDecode[64];
		// decode
		lil_u8 compI = mScanCompOrder[i];
		Component *pComp = &mComp[compI];
		lil_u16 *pQuant = mQt[pComp->quantI];
		HuffmanMsb *pHuffDc = mHuffDc[pComp->huffDcI];
		HuffmanMsb *pHuffAc = mHuffAc[pComp->huffAcI];
		for (lil_u32 dataUnitY = 0; dataUnitY < pComp->sampFactV; dataUnitY++) {
			for (lil_u32 dataUnitX = 0; dataUnitX < pComp->sampFactH; dataUnitX++) {
				// [Spec A.3.5 - Differential DC coding]
				lil_u32 huff = pHuffDc->JpegDecode(*mReader);
				if (huff == 0xffffffff)
					return LE_BADIMAGEDATA;

				lil_i32 val = pComp->pred;
				if (huff != 0)
					val += pHuffDc->JpegReceiveExtend(*mReader, huff);
				table[0] = val * pQuant[0];
				pComp->pred = val;

				// 1B) DECODE 63 AC
				// Yes, the spec uses variables "ssss" and "rrrr"
				memset(table + 1, 0, 63 * sizeof(table[0]));
				for (lil_u32 k = 1; k < 64; ) {
					huff = pHuffAc->JpegDecode(*mReader);
					if (huff == 0xffffffff)
						return LE_BADIMAGEDATA;
					lil_u32 ssss = huff & 15;
					//lil_u32 rrrr = huff >> 4;
					if (ssss == 0) {
						if (huff == 0xf0)  // rrrr == 15
							k += 16;
						else 
							break;
					}
					else {
						k += huff >> 4; // rrrr
						// [Spec A.3.4 - Quantization and dequantization]
						table[FROM_ZIG_ZAG[k]] = pQuant[k] * pHuffAc->JpegReceiveExtend(*mReader, ssss);
						k++;
					}
				}

				lil_u8 *pData = &pComp->data[(dataUnitY * 8 * pComp->sampFactH + dataUnitX) * 8];
				lil_u32 stride = 8 * pComp->sampFactH;
				if (mScanNumComp == 1) {
					pData = mStrip.GetPtr<lil_u8>() + mMcuX * 8;
					stride = mStripBpl;
				}

				// Record data unit into minimal coded unit for component
				IDCT(table, pData, stride);

				//lil_u8 *pIn = tableDecode;
				// Less memory accesses if we just copy this one time directly
				//for (lil_u8 y = 0; y < 8; y++) {
				//	for (lil_u8 x = 0; x < 8; x++)
				//		pData[x] = *pIn++;
				//	pData += stride;
				//}
				
			}
		}
		// 2) Dequantize - happened while reading data unit above
		// 3) IDCT
		// [Spec A.3.3 - FDCT And IDCT]
	}

	// Decode the MCU into the mcu strip now
	if (mScanNumComp == 3) {
		// Everything below depends on this...
		// NN algorithm.
		LIL_ASSERT(mCh == 4);
		for (lil_u32 i = 0; i < mScanNumComp; i++) {
			lil_u8 compI = mScanCompOrder[i];
			Component *pComp = &mComp[compI];
			lil_u32 w = 8 * pComp->sampFactH;
			lil_u32 xBytes = mMcuW * mCh;
			lil_u8 *pStrip = mStrip.GetPtr<lil_u8>() + mMcuX * xBytes;
			lil_u8 *pData = pComp->data;
			lil_u32 yMult = mComp[0].sampFactV / pComp->sampFactV;
			lil_u32 xMult = mComp[0].sampFactH / pComp->sampFactH;
			for (lil_u32 yi = 0; yi < mMcuH / yMult; yi++) {
				for (lil_u32 ym = 0; ym < yMult; ym++) {
					lil_u8 *pDataM = pData;
					if (xMult == 1) {
						for (lil_u32 xi = i; xi < xBytes;) {
							pStrip[xi] = *pDataM++;
							xi += 4;
						}
					}
					else if (xMult == 2) {
						for (lil_u32 xi = i; xi < xBytes;) {
							pStrip[xi] = *pDataM;
							pStrip[xi + 4] = *pDataM++;
							xi += 8;
						}
					}
					else if (xMult == 4) {
						for (lil_u32 xi = i; xi < xBytes;) {
							pStrip[xi] = *pDataM;
							pStrip[xi + 4] = *pDataM;
							pStrip[xi + 8] = *pDataM;
							pStrip[xi + 12] = *pDataM++;
							xi += 16;
						}
					}
					else { // When would this happen??
						for (lil_u32 xi = i; xi < xBytes;) {
							for (lil_u32 xm = 0; xm < xMult; xm++) {
								pStrip[xi] = *pDataM;
								xi += 4;
							}
							pDataM++;
						}
					}
					pStrip += mStripBpl;
				}
				pData += w;
			}
		}
		lil_u32 xBytes = mMcuW * mCh;
		lil_u8 *pStrip = mStrip.GetPtr<lil_u8>() + mMcuX * xBytes;
		for (lil_u32 yi = 0; yi < mMcuH; yi++) {
			lil_u8 *pOut = pStrip;
			for (lil_u32 xi = 0; xi < mMcuW; xi++) {
				YCbCrToRgb(pOut[0], pOut[1], pOut[2], pOut[2], pOut[1], pOut[0]);
				pOut[3] = 255;
				pOut += 4;
			}
			pStrip += mStripBpl;
		}
	}

	// [Spec .. - ]
	// Update restart counter
	mRestartCurr++;
	if (mRestartInterval && mRestartCurr >= mRestartInterval) {
		// Expect a reset marker here
		DecodeReset();
		lil_u32 nextVal = 0;
		mReader->ReadEntropyBits(nextVal, 16);
		if (nextVal < MARKER_RESTART_0 || nextVal > MARKER_RESTART_7) {
			mMcuY = mMcuCountY;
			mMcuX = mMcuCountX;
			mDecoding = false;
			if (nextVal == MARKER_END_OF_IMAGE) {
				mReader->UndoBits(nextVal, 16);
				return LE_NONE;
			}
			return LE_BADIMAGEDATA;
		}
	}

	// Track iteration
	mMcuX++;
	if (mMcuX == mMcuCountX) {
		mMcuX = 0;
		mMcuY++;
		if (mMcuY == mMcuCountY) {
			mDecoding = false;
			// [Spec B.1.1.5 - Markers in entropy coded data segments]
			mReader->AlignToByte();
		}
	}
	return LE_NONE;
}

} // lil
