#ifndef LILCODECRLE_H_INCLUDED
#define LILCODECRLE_H_INCLUDED

#include "Lil.h"
#include "LilMem.h"

namespace lil {

class RleEncoder; // Private class

//-----------------------------------------------------------------------------------------------------------
struct LilCodecRleBmp {
	LilCodecRleBmp(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat, bool bits4);
	~LilCodecRleBmp();
	
	// LilIoDesc who's read/write routines use this codec
	virtual const LilIoDesc *GetIoPtr() const { return &ioOutput; }

	// ENCODE AND DECODE
	// This will be the what goes into the codec (whether it's data to decode, or data to encode)
	const LilIoDesc *pIoInput;
	// Holds information like width and height... DONT use for image data - that will come
	// to/from memory pointers.
	const LilImageDesc *pImage;   
	lil_u32 shift;

	// Stream vars
	lil_i32 y, x;
	lil_u32 totalDecoded;
	lil_u32 bufSize;
	lil_u8 buf[256];  // Works with byte-size values, including counts

	// ENCODE
	RleEncoder *encoder;

	// This only matters for output, since the data may be in a different format (post-conversion)
	// to that of the original image where we get other info such as width and height.
	const LilFormatDesc *pFormat;

	// DECODE
	lil_u32 bytesPending;  // How many output bytes of the current operation are pending?
	bool ready;

private:
	// This has the result of the Codec.  Use the read function to get the result of encoding, and use
	// the write function to get the result of decoding.
	LilIoDesc ioOutput;
};

//-----------------------------------------------------------------------------------------------------------
struct LilCodecRleTga {
	LilCodecRleTga(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat);
	~LilCodecRleTga();
	
	// LilIoDesc who's read/write routines use this codec
	virtual const LilIoDesc *GetIoPtr() const { return &ioOutput; }

	// ENCODE AND DECODE
	// This will be the what goes into the codec (whether it's data to decode, or data to encode)
	const LilIoDesc *pIoInput;
	// Holds information like width and height... DONT use for image data - that will come
	// to/from memory pointers.
	const LilImageDesc *pImage;   
	lil_u32 pixBytes;  // number of bytes in a single pixel, could be 1 to 8 - depends on channel count
	lil_u32 pixByte; // Used during decode and encode to determine which byte is being read/written

	// Stream vars
	lil_u32 totalDecoded;
	lil_u32 bufSize;
	// Not much info will be stored here.  Holds up to 1 byte for command + 8 bytes (4 x 16 bit). 16 bytes for packing.
	lil_u8 buf[16];

	// ENCODE
	RleEncoder *encoder;

	// This only matters for output, since the data may be in a different format (post-conversion)
	// to that of the original image where we get other info such as width and height.
	const LilFormatDesc *pFormat;

	// DECODE
	bool ready;
	lil_u32 bytesPending;  // to be written (not necessarily what's in the buffer)

private:
	// This has the result of the Codec.  Use the read function to get the result of encoding, and use
	// the write function to get the result of decoding.
	LilIoDesc ioOutput;
};

//-----------------------------------------------------------------------------------------------------------
struct LilCodecPackbits {
	// If the byte counts aren't pre-image, then they will be pre-row.
	LilCodecPackbits(const LilIoDesc *pIoInput, const LilImageDesc *pImage, const LilFormatDesc *pFormat, 
		lil_u32 *pRowBytes, bool preImage);
	~LilCodecPackbits();
	
	// LilIoDesc who's read/write routines use this codec
	virtual const LilIoDesc *GetIoPtr() const { return &ioOutput; }

	// ENCODE AND DECODE
	// This will be the what goes into the codec (whether it's data to decode, or data to encode)
	const LilIoDesc *pIoInput;
	// Holds information like width and height... DONT use for image data - that will come
	// to/from memory pointers.
	const LilImageDesc *pImage;   

	// Stream vars
	lil_u32 totalEncoded;   // Num encoded bytes read when decoding/written when encoding
	lil_u32 totalDecoded;  // Num decoded bytes written when decoding/read when encoding
	lil_u32 bufSize;
	lil_i8 buf[2];  // Not much info will be stored here.  Holds up to 1 byte for command + 8 bytes (4 x 16 bit)
	LilBuffer rowBytes;
	lil_u32 *pRowBytes;
	lil_u32 currRowEnd;
	bool preImage;  // Location of each row's byte count - pre-image or pre-row
	LilBuffer rawBuf;
	lil_u32 rawBufPos;
	lil_u32 yPos;

	// ENCODE
	RleEncoder *encoder;
	lil_u32 resetAtZero;  // When this reaches 0, RESET and stop encoding - reached a plane boundary.
	lil_u32 rowByteStart;  // File pos for images that want a per-row byte count before the image starts

	// This only matters for output, since the data may be in a different format (post-conversion)
	// to that of the original image where we get other info such as width and height.
	const LilFormatDesc *pFormat;

	// DECODE
	bool ready;
	lil_u32 bytesPending;  // to be written (not necessarily what's in the buffer)

private:
	// This has the result of the Codec.  Use the read function to get the result of encoding, and use
	// the write function to get the result of decoding.
	LilIoDesc ioOutput;
};

}  // namespace lil

#endif // LILCODECRLE_H_INCLUDED