#include <algorithm>
#include "LilColorspace.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilUtilFormat.h"
#include "LilPlatform.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(Void) {
	LIL_UNUSED_PARAM(pSrcs);
	LIL_UNUSED_PARAM(pDests);
	LIL_UNUSED_PARAM(channelDepth);
	LIL_UNUSED_PARAM(srcHasAlpha);
	LIL_UNUSED_PARAM(destHasAlpha);
	LIL_UNUSED_PARAM(width);
	LIL_UNUSED_PARAM(planar);
}

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(RgbToLumin) {
	float rMult = globalConfig.coeffRgbToLumin[0];
	float gMult = globalConfig.coeffRgbToLumin[1];
	float bMult = globalConfig.coeffRgbToLumin[2];
	lil_u32 maxVal = (1u << channelDepth) - 1;
	if (planar)
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 r = pSrcs[0].ReadBits(channelDepth);
			lil_u32 g = pSrcs[1].ReadBits(channelDepth);
			lil_u32 b = pSrcs[2].ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs[3].ReadBits(channelDepth) : maxVal;
			pDests[0].WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests[1].WriteBits(channelDepth, a);
		}
	else
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 r = pSrcs->ReadBits(channelDepth);
			lil_u32 g = pSrcs->ReadBits(channelDepth);
			lil_u32 b = pSrcs->ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs->ReadBits(channelDepth) : maxVal;
			pDests->WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests->WriteBits(channelDepth, a);
		}
}

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(BgrToLumin) {
	float rMult = globalConfig.coeffRgbToLumin[0];
	float gMult = globalConfig.coeffRgbToLumin[1];
	float bMult = globalConfig.coeffRgbToLumin[2];
	lil_u32 maxVal = (1u << channelDepth) - 1;
	if (planar)
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 r = pSrcs[0].ReadBits(channelDepth);
			lil_u32 g = pSrcs[1].ReadBits(channelDepth);
			lil_u32 b = pSrcs[2].ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs[3].ReadBits(channelDepth) : maxVal;
			pDests[0].WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests[1].WriteBits(channelDepth, a);
		}
	else
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 b = pSrcs->ReadBits(channelDepth);
			lil_u32 g = pSrcs->ReadBits(channelDepth);
			lil_u32 r = pSrcs->ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs->ReadBits(channelDepth) : maxVal;
			pDests->WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests->WriteBits(channelDepth, a);
		}
}

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(ArgbToLumin) {
	LIL_ASSERT(srcHasAlpha);
	float rMult = globalConfig.coeffRgbToLumin[0];
	float gMult = globalConfig.coeffRgbToLumin[1];
	float bMult = globalConfig.coeffRgbToLumin[2];
	if (planar)
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 r = pSrcs[0].ReadBits(channelDepth);
			lil_u32 g = pSrcs[1].ReadBits(channelDepth);
			lil_u32 b = pSrcs[2].ReadBits(channelDepth);
			lil_u32 a = pSrcs[3].ReadBits(channelDepth);
			pDests[0].WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests[1].WriteBits(channelDepth, a);
		}
	else
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 a = pSrcs->ReadBits(channelDepth);
			lil_u32 r = pSrcs->ReadBits(channelDepth);
			lil_u32 g = pSrcs->ReadBits(channelDepth);
			lil_u32 b = pSrcs->ReadBits(channelDepth);
			pDests->WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests->WriteBits(channelDepth, a);
		}
}

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(AbgrToLumin) {
	LIL_ASSERT(srcHasAlpha);
	float rMult = globalConfig.coeffRgbToLumin[0];
	float gMult = globalConfig.coeffRgbToLumin[1];
	float bMult = globalConfig.coeffRgbToLumin[2];
	if (planar)
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 r = pSrcs[0].ReadBits(channelDepth);
			lil_u32 g = pSrcs[1].ReadBits(channelDepth);
			lil_u32 b = pSrcs[2].ReadBits(channelDepth);
			lil_u32 a = pSrcs[3].ReadBits(channelDepth);
			pDests[0].WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests[1].WriteBits(channelDepth, a);
		}
	else
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 a = pSrcs->ReadBits(channelDepth);
			lil_u32 b = pSrcs->ReadBits(channelDepth);
			lil_u32 g = pSrcs->ReadBits(channelDepth);
			lil_u32 r = pSrcs->ReadBits(channelDepth);
			pDests->WriteBits(channelDepth, static_cast<lil_u32>(r * rMult + g * gMult + b * bMult));
			if (destHasAlpha)
				pDests->WriteBits(channelDepth, a);
		}
}

//-----------------------------------------------------------------------------------------------------------
LIL_CONVERT_FUNC(CmykToRgb) {
	lil_u32 maxVal = (1u << channelDepth) - 1;
	if (planar)
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 c = pSrcs[0].ReadBits(channelDepth);
			lil_u32 m = pSrcs[1].ReadBits(channelDepth);
			lil_u32 y = pSrcs[2].ReadBits(channelDepth);
			lil_u32 k = pSrcs[3].ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs[4].ReadBits(channelDepth) : 0;
			lil_u32 r = 0, g = 0, b = 0;
			CmykToRgb(maxVal, maxVal, maxVal, c, m, y, k, r, g, b);
			pDests[0].WriteBits(channelDepth, r);
			pDests[1].WriteBits(channelDepth, g);
			pDests[2].WriteBits(channelDepth, b);
			if (destHasAlpha)
				pDests[3].WriteBits(channelDepth, a);
		}
	else
		for (lil_u32 i = 0; i < width; i++) {
			lil_u32 c = pSrcs->ReadBits(channelDepth);
			lil_u32 m = pSrcs->ReadBits(channelDepth);
			lil_u32 y = pSrcs->ReadBits(channelDepth);
			lil_u32 k = pSrcs->ReadBits(channelDepth);
			lil_u32 a = srcHasAlpha ? pSrcs->ReadBits(channelDepth) : 0;
			lil_u32 r = 0, g = 0, b = 0;
			CmykToRgb(maxVal, maxVal, maxVal, c, m, y, k, r, g, b);
			pDests->WriteBits(channelDepth, r);
			pDests->WriteBits(channelDepth, g);
			pDests->WriteBits(channelDepth, b);
			if (destHasAlpha)
				pDests->WriteBits(channelDepth, a);
		}
}

}  // namespace lil
