#ifndef LILCOLORSPACE_H_INCLUDED
#define LILCOLORSPACE_H_INCLUDED

#include "LilPlatform.h"

namespace lil {

class BitIoMsb;

//-----------------------------------------------------------------------------------------------------------
// YCBCR
// CCIR Recommendation 601 
// (this version is pretty specific to jpeg images)
LIL_INLINE void RgbToYCbCr(lil_i32 r, lil_i32 g, lil_i32 b, lil_u32 &y, lil_u32 &cb, lil_u32 &cr);
LIL_INLINE void YCbCrToRgb(lil_i32 y, lil_i32 cb, lil_i32 cr, lil_i32 &r, lil_i32 &g, lil_i32 &b);
LIL_INLINE void YCbCrToRgb(lil_i32 y, lil_i32 cb, lil_i32 cr, lil_u8 &r, lil_u8 &g, lil_u8 &b);
LIL_INLINE void RgbToCmy(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 r, lil_u32 g, lil_u32 b, lil_u32 &c, lil_u32 &m, lil_u32 &y);
LIL_INLINE void CmyToRgb(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 c, lil_u32 m, lil_u32 y, lil_u32 &r, lil_u32 &g, lil_u32 &b);
LIL_INLINE void RgbToCmyk(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 r, lil_u32 g, lil_u32 b, lil_u32 &c, lil_u32 &m, lil_u32 &y, lil_u32 &k);
LIL_INLINE void CmykToRgb(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 c, lil_u32 m, lil_u32 y, lil_u32 k, lil_u32 &r, lil_u32 &g, lil_u32 &b);

// This can be used both for declaration and definition.
#define LIL_CONVERT_FUNC(name) \
	void Convert##name(BitIoMsb *pSrcs, BitIoMsb *pDests, lil_u8 channelDepth, bool srcHasAlpha, bool destHasAlpha, \
		lil_u32 width, bool planar)

LIL_CONVERT_FUNC(Void);
LIL_CONVERT_FUNC(RgbToLumin);
LIL_CONVERT_FUNC(BgrToLumin);
LIL_CONVERT_FUNC(ArgbToLumin);
LIL_CONVERT_FUNC(AbgrToLumin);
LIL_CONVERT_FUNC(CmykToRgb);

//-----------------------------------------------------------------------------------------------------------
typedef void (* ConvertFunc)(BitIoMsb *, BitIoMsb *, lil_u8, bool, bool, lil_u32, bool);
const ConvertFunc CONVERT_FUNCTION[15][15] = {
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, ConvertVoid, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertRgbToLumin, ConvertRgbToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertBgrToLumin, ConvertBgrToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertArgbToLumin, ConvertArgbToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertAbgrToLumin, ConvertAbgrToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertRgbToLumin, ConvertRgbToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, ConvertBgrToLumin, ConvertBgrToLumin, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, ConvertCmykToRgb, 0, 0, 0, ConvertCmykToRgb, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, ConvertCmykToRgb, 0, 0, 0, ConvertCmykToRgb, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
};



//-----------------------------------------------------------------------------------------------------------
// INLINED IMPL.
//-----------------------------------------------------------------------------------------------------------
// YCBCR
// CCIR Recommendation 601 
// (this version is pretty specific to jpeg images)
LIL_INLINE void RgbToYCbCr(lil_i32 r, lil_i32 g, lil_i32 b, lil_u32 &y, lil_u32 &cb, lil_u32 &cr) {
	y = static_cast<lil_u32>(.299 * r + .587 * g + .114 * b);
	cb = static_cast<lil_u32>(-.1687 * r - .3313 * g + .5 * b + 128);
	cr = static_cast<lil_u32>(.5 * r - .4187 * g - .0813 * b + 128);
}

LIL_INLINE void YCbCrToRgb(lil_i32 y, lil_i32 cb, lil_i32 cr, lil_i32 &r, lil_i32 &g, lil_i32 &b) {
	// @@ Make it fixed point
	lil_i32 yFixed = LIL_INT_TO_FIXED(y);
	lil_i32 crm = cr - 128;
	lil_i32 cbm = cb - 128;
	const lil_i32 m1 = LIL_FLOAT_TO_FIXED(1.402);
	const lil_i32 m2 = LIL_FLOAT_TO_FIXED(.34414);
	const lil_i32 m3 = LIL_FLOAT_TO_FIXED(.71414);
	const lil_i32 m4 = LIL_FLOAT_TO_FIXED(1.772);
/*	r = LIL_FIXED_TO_INT(yFixed + m1 * crm);
	g = LIL_FIXED_TO_INT(yFixed - m2 * cbm - m3 * crm);
	b = LIL_FIXED_TO_INT(yFixed + m4 * cbm);
	lil_u32 t;
	// Trick to clamp
	if (t = r >> 8) r = ~t >> 24;
	if (t = g >> 8) g = ~t >> 24;
	if (t = b >> 8) b = ~t >> 24;*/
	r = yFixed + m1 * crm;
	g = yFixed - m2 * cbm - m3 * crm;
	b = yFixed + m4 * cbm;
	lil_u32 t;
	t = r >> 24;
	if (t) r = ~t >> 24;
	else r >>= 16;
	t = g >> 24;
	if (t) g = ~t >> 24;
	else g >>= 16;
	t = b >> 24;
	if (t) b = ~t >> 24;
	else b >>= 16;
}

LIL_INLINE void YCbCrToRgb(lil_i32 y, lil_i32 cb, lil_i32 cr, lil_u8 &r, lil_u8 &g, lil_u8 &b) {
	// @@ Make it fixed point
	lil_i32 yFixed = LIL_INT_TO_FIXED(y);
	lil_i32 crm = cr - 128;
	lil_i32 cbm = cb - 128;
	const lil_i32 m1 = LIL_FLOAT_TO_FIXED(1.402);
	const lil_i32 m2 = LIL_FLOAT_TO_FIXED(.34414);
	const lil_i32 m3 = LIL_FLOAT_TO_FIXED(.71414);
	const lil_i32 m4 = LIL_FLOAT_TO_FIXED(1.772);
	lil_u32 t;
	lil_i32 c = yFixed + m1 * crm;
	t = c >> 24;
	if (t) r = ~t >> 24;
	else r = static_cast<lil_u8>(c >> 16);
	c = yFixed - m2 * cbm - m3 * crm;
	t = c >> 24;
	if (t) g = ~t >> 24;
	else g = static_cast<lil_u8>(c >> 16);
	c = yFixed + m4 * cbm;
	t = c >> 24;
	if (t) b = ~t >> 24;
	else b = static_cast<lil_u8>(c >> 16);
}

//-----------------------------------------------------------------------------------------------------------
// Y (LUMINANCE)
//LIL_INLINE void RgbToY(lil_u32 r, lil_u32 g, lil_u32 b, lil_u32 &y) {
	//y = static_cast<lil_u32>(.2126 * r + .7152 * g + .0722 * b);
//}
// ConvertYToRgb can be done within normal functionality

//-----------------------------------------------------------------------------------------------------------
// CMYK
// Naive CMYK conversion
LIL_INLINE void RgbToCmy(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 r, lil_u32 g, lil_u32 b, lil_u32 &c, lil_u32 &m, lil_u32 &y) {
	c = maxR - r;
	m = maxG - g;
	y = maxB - b;
}

LIL_INLINE void CmyToRgb(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 c, lil_u32 m, lil_u32 y, lil_u32 &r, lil_u32 &g, lil_u32 &b) {
	r = maxR - c;
	g = maxG - m;
	b = maxB - y;
}

LIL_INLINE void RgbToCmyk(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 r, lil_u32 g, lil_u32 b, lil_u32 &c, lil_u32 &m, lil_u32 &y, lil_u32 &k) {
	k = GetMin(r, g, b);
	c = maxR - r - k;
	m = maxG - g - k;
	y = maxB - b - k;
}

LIL_INLINE void CmykToRgb(lil_u32 maxR, lil_u32 maxG, lil_u32 maxB, lil_u32 c, lil_u32 m, lil_u32 y, lil_u32 k, lil_u32 &r, lil_u32 &g, lil_u32 &b) {
	r = maxR - GetMin(maxR, c + k);
	g = maxG - GetMin(maxG, m + k);
	b = maxB - GetMin(maxB, y + k);
}

}  // namespace lil

#endif  // LILCOLORSPACE_H_INCLUDED
