#ifndef LILCONVERT_H_INCLUDED
#define LILCONVERT_H_INCLUDED

#include "Lil.h"
#include "LilPlatform.h"

namespace lil {

class LilBuffer;

// Fills a channel with a single constant value, which is pointed to by pSrcData.
LilErrorType FillChannel(const void *pSrcData, void *pDestData, const LilFormatDesc *pDestFormat,
	lil_u32 destCh, lil_u32 width, lil_u32 height);

// Fills alpha channel specifically with all '1' bits.
LilErrorType FillAlpha(void *pDestData, LilImageDesc *pImage);

LilErrorType ConvertChannelToChannel(const void *pSrcData, const LilFormatDesc *pSrcFormat, lil_u32 srcCh,
	void *pDestData, const LilFormatDesc *pDestFormat, lil_u32 destCh, lil_u32 width, lil_u32 height);

// Run all data that will be used with ConvertData through this function first!!
// It ensures everything is the proper endianness - this can get quite complicated
// with the various possible formats.
void MakeBigEndian(void *data, lil_u32 width, lil_u32 height, const LilFormatDesc *format);

// Convert the data in one block of memory into another block of memory.
// Note that both of these MUST have the properly allocated memory!
// Specifying bpl allows us to write out portions or sub-rectangles of one image to another.
LilErrorType ConvertData(const LilMemDesc *pMemFuncs,
	const void *pSrcData, const LilFormatDesc *pSrcFormat, const LilPaletteDesc *pSrcPalette,
	void *pDestData, const LilFormatDesc *pDestFormat, LilPaletteDesc *pDestPalette,
	lil_u32 width, lil_u32 height, bool isLittleEndian);

LilErrorType ConvertPalette(const LilImageDesc *pImage, LilPaletteDesc *pDest, LilBuffer *pTempBuf);

}  // namespace lil

#endif // LILCONVERT_H_INCLUDED
