#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_JPG

#include "LilConvert.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilIo.h"
#include "LilUtilFormat.h"
#include "LilCodecJpeg.h"
#include "jpge.h"

namespace {

using namespace lil;

}



namespace lil {

//-----------------------------------------------------------------------------------------------------------
bool IsExtensionJpg(const char *fileName)
{
	if (CheckExtension(fileName, ".JPG") || CheckExtension(fileName, ".JPEG") ||
		CheckExtension(fileName, ".JPE") || CheckExtension(fileName, ".JFIF"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableJpg(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);
	lil_u16 soi = Read16be(io);
	lil_u16 app0 = Read16be(io);
	Read16be(io);
	lil_u32 jfif = Read32be(io);
	lil_u8 zero = Read8(io);

	//if (soi == MARKER_START_OF_IMAGE && app0 == MARKER_APP0 && jfif == ID_JFIF && zero == ZERO_8)
	//	return true;
	// Should be able to read jpegs even if NOT jfif...... right?
	if (soi == MARKER_START_OF_IMAGE)
        return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadJpg(const LilIoDesc *io, LilImageDesc *image, LilLoadType request,
								 const LilFormatDesc *imageFormat)
{
	// Jpegs especially benefit from loading straight from memory. Should this apply to all
	// formats? Or at least heavily compressed ones that won't take up a whole lot of memory?
	// If we can guarantee jpegs will always go into memory first, we can use much faster
	// functions instead of the plug-in IO functions, which the compiler can't possibly
	// inline...
	lil_u64 size = GetIoSize(io);
	LilBuffer buf;
	if (!buf.Realloc(&image->memFuncs, size))
		return LE_NOTENOUGHMEMORY;
	io->readFunc(io->handle, size, buf.GetPtr<lil_u8>());

	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;
//	lil_u8 zero = 0, units = 0, mThumbW = 0, mThumbH = 0, code = 0;
//	lil_u16 app0 = 0, headerLen = 0, version = 0, xUnits = 0, yUnits = 0, marker = 0;
//	lil_u32 id = 0;
//	bool done = false;

	LilCodecJpeg jpeg(buf.GetIoDescPtr(), image, imageFormat ? imageFormat : &image->format);

	io->seekFunc(io->handle, 0, LS_SET);

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	err = jpeg.DecodeStart();
	if (err != LE_NONE)
		return err;

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	// 2c) READ PALETTE INFO
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	err = jpeg.DecodeImageInfo();
	if (err != LE_NONE)
		return err;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	return ReadImageData(jpeg.GetIoOutputPtr(), image, imageFormat, false, false);
}

class io_stream : public jpge::output_stream
{
	const LilIoDesc *pIo;
	LilErrorType err;

public:
   io_stream(const LilIoDesc *pIo) : pIo(pIo), err(LE_NONE) { }
   virtual bool put_buf(const void* pBuf, int len)
   {
	   err = pIo->writeFunc(pIo->handle, len, pBuf);
	   return err == LE_NONE;
   }
};

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveJpg(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	LilBuffer buf;
	if (!buf.Realloc(&image->memFuncs, CalcByteSize(&image->format, image->data.width, image->data.height)))
		return LE_NOTENOUGHMEMORY;
	image->memFuncs.imageReadFunc(image->data.data, buf.GetPtr<void>(), 0, 0, image->data.width, image->data.height, 0, CalcBpl(&image->format, image->data.width));

	io_stream jio(io);
	jpge::jpeg_encoder dst_image;
	if (!dst_image.init(&jio, image->data.width, image->data.height, image->format.numChannels))
		return LE_UNSUPPORTEDFILEFEATURE;

	for (int pass_index = 0; pass_index < dst_image.get_total_passes(); pass_index++) {
		for (int i = 0; i < image->data.height; i++) {
			const lil_u8 *pBuf = buf.GetPtr<lil_u8>() + i * image->data.width * image->format.numChannels;
			if (!dst_image.process_scanline(pBuf))
				return LE_BADCOMPRESSION;
		}
		if (!dst_image.process_scanline(NULL))
			return LE_BADCOMPRESSION;
	}

	dst_image.deinit();

	return LE_NONE;

	/*
	LilErrorType err = LE_NONE;

	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	const LilFormatDesc *pOutFormat = imageFormat ? imageFormat : &image->format;
	lil_u16 bpp = static_cast<lil_u16>(CalcBpp(pOutFormat));
	lil_u32 imageSize = CalcByteSize(pOutFormat, image->data.width, image->data.height);
	bool detectComp = image->extended.compression == LC_ANY;
	lil_u16 headerLen = 0;
	LilCompressionType compression = detectComp ? LC_BMP_NONE : image->extended.compression;

	LIL_UNUSED_PARAM(err);
	LIL_UNUSED_PARAM(bpp);
	LIL_UNUSED_PARAM(imageSize);
	LIL_UNUSED_PARAM(compression);

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	// Check format that we will be saving.  Compensate for any possibility to use LC_BMP_BITFIELDS.
/*	LilFormatDesc compareFormat;
	err = SetFormatFromBppWin(&compareFormat, static_cast<lil_u8>(CalcBpp(pOutFormat)));
	if (err != LE_NONE)
		return err;
	compareFormat.flipped = pOutFormat->flipped;
	if (IsFormatDifferent(&compareFormat, pOutFormat, 0)) {
		lil_u32 add4 = 0, add3 = 0;
		for (int i = 0; i < 3; i++)
			add3 += pOutFormat->bitCounts[i];
		add4 = add3 + pOutFormat->bitCounts[3];
		if ((detectComp || compression == LC_BMP_BITFIELDS) &&
			pOutFormat->planar == false && pOutFormat->blockLineByteAlign == 4 &&
			(bpp == 16 || bpp == 32) &&
			(pOutFormat->dataFormat == LDF_SIMPLEBITS &&
				((pOutFormat->numChannels == 4 && pOutFormat->channelOrder == LCO_ARGB && add4 == bpp) ||
					(pOutFormat->numChannels == 3 && pOutFormat->channelOrder == LCO_RGB && add3 == bpp)))
			)
		{
			compression = LC_BMP_BITFIELDS;
		}
		else
			return LE_UNSUPPORTEDFORMAT;
	}

	if (detectComp) {
		if (bpp == 8)
			compression = LC_BMP_RLE8;
		if (bpp == 4)
			compression = LC_BMP_RLE4;
	}

	if (colors) {
		LilFormatDesc palFormat;
		SetFormat(&palFormat, LDF_SIMPLE, 8, 3, 4, false, LCO_BGR, false);
		if (!IsFormatDifferent(&image->palette.format, &palFormat, IFD_IGNORE_FLIPPED | IFD_IGNORE_PLANAR | IFD_IGNORE_STRIDE)) {
			palUseType = 1;
			headerSize = HEADER_SIZE_V1;
			palChans = 3;
			//headerType = ID_OS2_BMP;
		}
		SetFormat(&palFormat, LDF_SIMPLE, 8, 4, 4, false, LCO_BGRA, false);
		if (!IsFormatDifferent(&image->palette.format, &palFormat, IFD_IGNORE_FLIPPED | IFD_IGNORE_PLANAR | IFD_IGNORE_STRIDE))
			palUseType = 2;
		if (!palUseType)
			return LE_UNSUPPORTEDPALETTEFORMAT;
		// @@ If it is a 3 channel palette, we can only save as OS2 bitmap, if that's even possible.  Should check
		// that it is a compatible OS2 type (only 1, 4, 8, 24 mBits!!!)
	}

	// Needed all previous information to calculate these
	lil_u32 dataOffset = 14 + headerSize + colors * palChans + (compression == LC_BMP_BITFIELDS ? 12 : 0);
	lil_u32 fileSize = dataOffset + imageSize;
*/
	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
/*	io->seekFunc(io->handle, 0, LS_SET);

	headerLen = 16;
	WriteFbe(io, "22241", MARKER_START_OF_IMAGE, MARKER_APP0, headerLen, ID_JFIF, ZERO_8);
	lil_u16 one = 1;
	WriteFbe(io, "2122 11", SUPPORTED_VERSION, UNITS_ASPECT_RATIO, one, one, ZERO_8, ZERO_8);

	// Write tables/misc
	Write16be(io, MARKER_QUANTIZATION);
	Write16be(io, MARKER_HUFFMAN_SPEC);
	// Write Start of Frame for compression method
	Write16be(io, MARKER_BASELINE_DCT);
	// Rep Point
	// Write Start of Scans
	Write16be(io, MARKER_START_OF_SCAN);
	// Write tabels/misc
	// Repeat until Image Finished
	Write16be(io, MARKER_END_OF_IMAGE);

	//------------------------
	// 3b) WRITE IMAGE INFO
/*	if (palUseType == 1) {
		// OS2 has 16 bit fields for w and h
		lil_u16 w = static_cast<lil_u16>(image->data.width);
		lil_u16 h = static_cast<lil_u16>(image->data.height);
		WriteFle(io, "2222", w, h, numPlanes, bpp);
	}
	else {
		WriteFle(io, "4422 444444", image->data.width, pOutFormat->flipped ? height : -height,
			numPlanes, bpp, compression - 1, imageSize,
			image->extended.fileFormat.bmp.ppmX, image->extended.fileFormat.bmp.ppmY,
			colors, importantColors);
		if (compression == LC_BMP_BITFIELDS) {
			lil_u32 rMask = 0, gMask = 0, bMask = 0;
			lil_u32 addC = pOutFormat->numChannels == 4 ? 1 : 0;
			bMask = (1 << pOutFormat->bitCounts[2 + addC]) - 1;
			gMask = ((1 << pOutFormat->bitCounts[1 + addC]) - 1) << pOutFormat->bitCounts[2 + addC];
			rMask = ((1 << pOutFormat->bitCounts[0 + addC]) - 1) << (pOutFormat->bitCounts[2 + addC] + pOutFormat->bitCounts[1 + addC]);
			WriteFle(io, "444", rMask, gMask, bMask);
		}
	}

	//------------------------
	// 3c) WRITE PALETTE INFO
	WriteN(io, colors * palChans, image->palette.lut);

	//------------------------
	// 4) SAVE IMAGE DATA
	std::auto_ptr<LilCodecRleBmp> codec;
	const LilIoDesc *pIo = io;
	if (compression == LC_BMP_RLE8 || compression == LC_BMP_RLE4) {
		codec = std::auto_ptr<LilCodecRleBmp>(new LilCodecRleBmp(io, image, &outFormat, compression == LC_BMP_RLE4));
		pIo = codec->GetIoPtr();
	}

	WriteImageData(pIo, image, &outFormat);
	*/

	//return LE_NONE;
}

}  // namespace lil

#endif // LIL_SUPPORTS_JPG
