#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_PCT

#include <memory>
#include "LilCodecRle.h"
#include "LilUtilFormat.h"
#include "LilMem.h"
#include "LilConvert.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
namespace {

lil_i32 OPDATA_SIZE_V1[256] = {
	0, -1, 8, 2, 1, 2, 4, 4, 2, 8, 8, 4, 4, 2, 4, 4,
	8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	8, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	8, 8, 8, 8, 8, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1,
	8, 8, 8, 8, 8, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1,
	8, 8, 8, 8, 8, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1,
	12, 12, 12, 12, 12, -1, -1, -1, 4, 4, 4, 4, 4, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0
};

lil_i32 OPDATA_SIZE_V2[256] = {
	0, -1, 8, 2, 1, 2, 4, 4, 2, 8, 8, 4, 4, 2, 4, 4,
	8, 1, -1, -1, -1, 2, 2, -1, -1, -1, 6, 6, 0, 6, 0, 6,
	8, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, 10, 8, -1,
	8, 8, 8, 8, 8, 8, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0,
	8, 8, 8, 8, 8, 8, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0,
	8, 8, 8, 8, 8, 8, 8, 8, 0, 0, 0, 0, 0, 0, 0, 0,
	12, 12, 12, 12, 12, 12, 12, 12, 4, 4, 4, 4, 4, 4, 4, 4,
	-1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
	-1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 2
};

const lil_u16 OP_CLIPRECT = 0x0001;
const lil_u16 OP_ENDPIC = 0x00ff;
const lil_u16 OP_HEADER = 0x0c00;
const lil_u16 OP_VERSION = 0x02ff;  // This one is really special - it's 2 bytes data, NOT 4!!

const lil_u16 OP_DATA_BITS_RECT = 0x0090;  // v1, v2
const lil_u16 OP_DATA_BITS_RGN = 0x0091;  // v1, v2
const lil_u16 OP_DATA_PACKBITS_RECT = 0x0098;  // v1, v2
const lil_u16 OP_DATA_PACKBITS_RGN = 0x0099;  // v1, v2
const lil_u16 OP_DATA_DIRECTBITS_RECT = 0x009a;  // v2
const lil_u16 OP_DATA_DIRECTBITS_RGN = 0x009b;  // v2
const lil_u16 OP_DATA_QT_COMP = 0x8200;  // v2
const lil_u16 OP_DATA_QT_RAW = 0x8201;  // v2

// IDs
const lil_u16 ID_V1 = 0x1101;
const lil_u16 ID_V2 = 0x0011;
const lil_u16 ID_V2_DATA = 0x02ff;
const lil_u16 EXT_NO = 0xffff;
const lil_u16 EXT_YES = 0xfffe;
const lil_u16 PIXTYPE_MONO = 0;
const lil_u16 PIXTYPE_RGB = 16;

const lil_u16 PACKTYPE_DEFAULT = 0;
const lil_u16 PACKTYPE_NONE = 1;
const lil_u16 PACKTYPE_24_OF_32 = 2;
const lil_u16 PACKTYPE_PACKBITS_16 = 3;
const lil_u16 PACKTYPE_PACKBITS_24 = 4;  // even though file should say 32 bpp... it's really 24

const lil_u16 ARITH_SRCCOPY = 0;

// Compressions
const lil_i32 GOOD_C[] = { LC_PCT_NONE, LC_PCT_PACKBITS };
const lil_u8 N_GOOD_C = LIL_ARRAY_SIZE(GOOD_C);
const lil_i32 BAD_C[] = { LC_PCT_JPEG };
const lil_u8 N_BAD_C = LIL_ARRAY_SIZE(BAD_C);

// Formats
// 32 bit planar means packed, non planar means raw
const LilFormatRange ranges[] = {
	{ LDF_SIMPLE, { {1}, {2}, {4}, {8} }, 4, 1, 1, FLAG_NO, LCO_SINGLE, FLAG_NO },
	{ LDF_SIMPLE, { {8} }, 1, 3, 1, FLAG_NO, LCO_RGB, FLAG_NO },
	{ LDF_SIMPLE, { {8} }, 1, 4, 1, FLAG_NO, LCO_RGBA, FLAG_NO },
	{ LDF_SIMPLEBITS, { {1, 5, 5, 5} }, 1, 4, 1, FLAG_NO, LCO_RGBA, FLAG_NO }
};
const lil_u8 N_RANGES = LIL_ARRAY_SIZE(ranges);

//-----------------------------------------------------------------------------------------------------------
bool IsValidId(lil_u16 id, lil_u32 ver2, lil_u16 ext, LilErrorType *pErr) {
	if (id == ID_V1) {
		if (pErr)
			*pErr = LE_NONE; //LE_UNSUPPORTEDFILEFEATURE;
		return false;
	}
	// If v2 "version opcode", check data (0x02, plus 0xff for padding), and v2 header op,
	// and whether it's either "v2" or "extended v2".
	else if (id == ID_V2 && (ver2 == ((ID_V2_DATA << 16) | OP_HEADER)) && (ext == EXT_NO || ext == EXT_YES || ext == 0)) {
		if (pErr)
			*pErr = LE_NONE;
	}
	else {
		if (pErr)
			*pErr = LE_BADFILEFORMAT;
		return false;
	}
	return true;
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u16 GetOp(const LilIoDesc *io, bool v2) {
	if (v2)
		return Read16be(io);
	return Read8(io);
}

//-----------------------------------------------------------------------------------------------------------
LIL_INLINE lil_u32 GetDataSize(const LilIoDesc *io, lil_u16 op, bool v2) {
	if (v2) {
		if (op < 256)
			return OPDATA_SIZE_V2[op];
		else if ((op >> 8) < 0x80)
			return (op >> 8) * 2;  // Typically this is the pattern, except for what follows
		else if ((op >> 8) == 0x80)
			return 0;
		else
			return Read32be(io);  // "4 + len", but we just read the 4 part
	}
	return OPDATA_SIZE_V1[op];
}

} // namespace



//-----------------------------------------------------------------------------------------------------------
bool IsExtensionPct(const char *fileName)
{
	if (CheckExtension(fileName, ".pct") || CheckExtension(fileName, ".pict"))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectablePct(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 522, LS_SET);
	lil_u16 id = Read16be(io);
	lil_u32 ver2 = Read32be(io);
	lil_u16 ext = Read16be(io);
	io->seekFunc(io->handle, 0, LS_SET);
	return IsValidId(id, ver2, ext, 0);
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadPct(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE READ
	LilErrorType err = LE_NONE;

	bool v2 = false, done = false, isDirect = false;
	lil_u16 id = 0, ext = 0, op = 0, sizeV1 = 0, t = 0, l = 0, b = 0, r = 0;
	lil_u32 ver2 = 0, dataSize = 0, dpiX = 0, dpiY = 0;
	LilCompressionType compression = LC_PCT_NONE;
	lil_u16 bpl = 0, dummy16 = 0, pixMapVer = 0, packType = 0;
	lil_u16 pixType = 0, bpp = 0, ch = 0, bits = 0;
	lil_u32 packSize = 0, planeOffset = 0, dummy32 = 0;
	LilChannelOrderType lco = LCO_SINGLE;
	lil_u8 align = 0;

	//------------------------
	// 2a) READ FILE FORMAT HEADER
	io->seekFunc(io->handle, 512, LS_SET);
	ReadFbe(io, "22222 2", &sizeV1, &t, &l, &b, &r, &id);
	if (id == ID_V2)
		ReadFbe(io, "42", &ver2, &ext);

	// Validate
	if (!IsValidId(id, ver2, ext, &err))
		return err;

	// Finish reading header if validated
	if (id == ID_V2) {
		v2 = true;
		if (ext == EXT_YES) {
			Read16be(io);
			dpiX = Read32be(io) >> 16;  // 16.16 fixed point
			dpiY = Read32be(io) >> 16;  // 16.16 fixed point
			io->seekFunc(io->handle, 12, LS_RELATIVE);  // I think official docs on this are 2 bytes off!
		}
		else  // How to decide which bounding rect to use??
			io->seekFunc(io->handle, 22, LS_RELATIVE);
	}

	//------------------------
	// 2b) READ SELECTED IMAGE INFO
	// 2c) READ PALETTE INFO
	// 3) SET IMAGE, FORMAT, EXTENDED INFO
	while (!done) {
		op = GetOp(io, v2);
		dataSize = GetDataSize(io, op, v2);
		switch (op) {
		// --- Binary image data! ---
		case OP_DATA_DIRECTBITS_RECT:  // 16 or 32
			isDirect = true;
			io->seekFunc(io->handle, 4, LS_RELATIVE); // Skip "baseAddr" field
		case OP_DATA_BITS_RECT:  // 1 2 4 or 8
		case OP_DATA_PACKBITS_RECT:  // 1 2 4 or 8
			ReadFbe(io, "2 2222 224 44 2222 444", &bpl, &t, &l, &b, &r, &pixMapVer,
				&packType, &packSize, &dpiX, &dpiY, &pixType, &bpp, &ch, &bits,
				&planeOffset, &dummy32, &dummy32);
			// Spec says default (0 means default) for 16/32 bit is 3/4 packType
			if (packType == PACKTYPE_DEFAULT && bpp == 32)
				packType = PACKTYPE_PACKBITS_24;
			switch (packType) {
				case PACKTYPE_DEFAULT: case PACKTYPE_PACKBITS_16: compression = LC_PCT_PACKBITS; break;
				case PACKTYPE_PACKBITS_24:
					compression = LC_PCT_PACKBITS;
				case PACKTYPE_24_OF_32:
					ch = 3;
					bpp = 24;
					if (bpl & 3)
						return LE_BADCOMPRESSION;
					bpl = (bpl & 0x8000) | (bpl - (bpl & 0x7fff) / 4);
					break;
				case PACKTYPE_NONE: break;
				default: return LE_UNSUPPORTEDCOMPRESSION;
			}
			// This just helps indicate if it's 1bpp or not
			if (bpl & 0x8000) {
				if (bpp == 1)
					return LE_BADFILEFORMAT;
				bpl &= 0x7fff;
				// bytes needed to represent a row
				lil_u32 total = (((r - l) * bpp) + 7) / 8;
				if (bpl == total)
					align = 1;
				else if (bpl == RoundUpToNearest(total, 2))
					align = 2;
				else if (bpl == RoundUpToNearest(total, 4))
					align = 4;
				else
					return LE_UNSUPPORTEDFORMAT;
			}
			else if (bpp != 1)
				return LE_BADFILEFORMAT;

			// Planar data is not supported!  planeOffset will always be 0.
			// NOTE: It looks like some files use this?
			if (planeOffset != 0) {
				ch = static_cast<lil_u16>(planeOffset) / bits;
				bpp = static_cast<lil_u16>(planeOffset);
				if (ch * bits != static_cast<lil_u16>(planeOffset))
					return LE_UNSUPPORTEDFILEFEATURE;
			}

			if (pixType == PIXTYPE_RGB) {
				if (!isDirect)
					return LE_BADFILEFORMAT;
				if ((ch != 3 && ch != 4) || (bpp != 16 && bpp != 24 && bpp != 32) || (bits != 8 && bits != 5))
					return LE_BADCOLORSPACE;
				lco = (ch == 3) ? LCO_RGB : LCO_ARGB;
			}
			else {
				if (pixType != PIXTYPE_MONO || isDirect)
					return LE_BADFILEFORMAT;
				if (ch != 1 || (bits != 1 && bits != 2 && bits != 4 && bits != 8) || bpp != bits)
					return LE_BADCOLORSPACE;
			}

			if (!SetImageSize(&image->data, r - l, b - t, &err))
				return err;
			if (!SetFormat(&image->format, LDF_SIMPLE, static_cast<lil_u8>(bits),
				static_cast<lil_u8>(ch), align, false, lco, false, &err))
				return err;
			if (!SetExtendedInfo(&image->extended, 1, compression, LD_INCHES, dpiX >> 16, dpiY >> 16, &err))
				return err;

			if (!isDirect) {
				// (4) unique id for color table, (2) flags
				io->seekFunc(io->handle, 6, LS_RELATIVE);
				lil_u16 size = Read16be(io) + 1;  // Always one less
				err = image->memFuncs.memReallocFunc(&image->palette.lut, size * 6);
				if (err != LE_NONE)
					return err;
				lil_u16 *pPal = reinterpret_cast<lil_u16 *>(image->palette.lut);
				image->palette.numEntries = static_cast<lil_u32>(size);
				if (!SetFormat(&image->palette.format, LDF_SIMPLE, 16, 3, 1, false, LCO_RGB, false, &err))
					return err;
				for (lil_u32 i = 0; i < size; i++) {
					ReadFbe(io, "2222", &dummy16, &pPal[0], &pPal[1], &pPal[2]);
					pPal += 3;
				}
			}

			// (8) Src rect, (8) dest rect, (2) arith transfer mode
			io->seekFunc(io->handle, 18, LS_RELATIVE);

			// @@@@ Possibly has a mask region next

			done = true;

			break;
		// @@ No pict files with rgns to test this with!
		case OP_DATA_DIRECTBITS_RGN:  // 16 or 32
		case OP_DATA_BITS_RGN:  // 1 2 4 or 8
		case OP_DATA_PACKBITS_RGN:  // 1 2 4 or 8
		case OP_DATA_QT_COMP:
		case OP_DATA_QT_RAW:
			compression = LC_PCT_JPEG;
			io->seekFunc(io->handle, dataSize, LS_RELATIVE);
			break;
		// --- Other stuff ---
		case OP_VERSION:  // Special case that doesn't follow rules for size determination
			io->seekFunc(io->handle, 2, LS_RELATIVE);
			break;
		case OP_ENDPIC:
			io->seekFunc(io->handle, dataSize, LS_RELATIVE);
			done = true;
			break;
		default:
			// This typically works, and when it doesn't, we should handle specially
			// above in a real case statement.
			if (dataSize == -1)
				dataSize = Read16be(io) - 2;  // includes the 2 bytes we just read!!
			io->seekFunc(io->handle, dataSize, LS_RELATIVE);
			break;
		}
	}

	// UGH Bitfields compression is kind of weird between different bmp files
	if (!CheckValue(compression, GOOD_C, N_GOOD_C, BAD_C, N_BAD_C, LE_UNSUPPORTEDCOMPRESSION, LE_BADCOMPRESSION, &err))
		return err;

	//------------------------
	// 4) LOAD IMAGE DATA
	if (request == LL_INFO)
		return LE_NONE;

	std::auto_ptr<LilCodecPackbits> codec;
	const LilIoDesc *pIo = io;
	LilBuffer buf;
	if (!buf.Realloc(&image->memFuncs, image->data.height * ch * 4))
		return LE_NOTENOUGHMEMORY;
	if (compression == LC_PCT_PACKBITS) {
		codec = std::auto_ptr<LilCodecPackbits>(new LilCodecPackbits(io, image,
			&image->format, buf.GetPtr<lil_u32>(), false));
		pIo = codec->GetIoPtr();
	}

	return ReadImageData(pIo, image, imageFormat, false, true);
}

/*---------------------------------------------------------------------------------------------------------*/
LilErrorType SavePct(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	//------------------------
	// 1) VARIABLES TO BE WRITTEN
	LilErrorType err = LE_NONE;

	LilFormatDesc outFormat = imageFormat ? *imageFormat : image->format;
	lil_u32 bpl = 0;
	lil_u16 bpp = static_cast<lil_u16>(CalcBpp(&outFormat));
	bool detectComp = image->extended.compression == LC_ANY;
	LilCompressionType compression = detectComp ? LC_PCT_PACKBITS : image->extended.compression;
	lil_u32 dpiX = image->extended.dpuX, dpiY = image->extended.dpuY;

	lil_u16 w = static_cast<lil_u16>(image->data.width), h = static_cast<lil_u16>(image->data.height);
	lil_u16 imageType = OP_DATA_PACKBITS_RECT, packType = PACKTYPE_DEFAULT, pixType = PIXTYPE_MONO;
	lil_u16 ch = 0;

	//------------------------
	// 2) GET IMAGE, FORMAT, EXTENDED INFO
	if (!MatchFormat(&outFormat, ranges, N_RANGES, 0, 0, &err))
		return err;
	bpl = CalcBpl(&outFormat, image->data.width);
	ch = static_cast<lil_u16>(outFormat.numChannels);

	if (image->data.width > 65535 || image->data.height > 65535)
		return LE_UNSUPPORTEDSIZE;

	//------------------------
	// 3a) WRITE FILE FORMAT HEADER
	// 3b) WRITE IMAGE INFO
	// 3c) WRITE PALETTE INFO
	io->seekFunc(io->handle, 512, LS_SET);
	// Basic header
	WriteFbe(io, "22222 2 22", ZERO_16, ZERO_16, ZERO_16, h, w, ID_V2, ID_V2_DATA, OP_HEADER);

	// V2 (NOT ext) header
	WriteFbe(io, "22 4 4 22 22 4", EXT_NO, EXT_NO, ZERO_32, ZERO_32, w, ZERO_16, h, ZERO_16, ZERO_32);

	// V2 clip region
	WriteFbe(io, "2 22222", OP_CLIPRECT, static_cast<lil_u16>(10), ZERO_16, ZERO_16, h, w);

	// V2 Image header
	if (bpp == 16 || bpp == 24 || bpp == 32) {
		imageType = OP_DATA_DIRECTBITS_RECT;
		if (compression == LC_PCT_PACKBITS) {
			if (bpp == 32) {
				if (detectComp)
					compression = LC_PCT_NONE;
				else
					return LE_UNSUPPORTEDFORMAT;
			}
			packType = (bpp == 16) ? PACKTYPE_PACKBITS_16 : PACKTYPE_PACKBITS_24;
		}
		if (compression == LC_PCT_NONE)
			packType = (bpp == 16) ? PACKTYPE_NONE : (bpp == 24 ? PACKTYPE_24_OF_32 : PACKTYPE_NONE);
		pixType = PIXTYPE_RGB;  // required ALWAYS for directbits

		// This will need some adjustments to match pct format
		if (bpp == 24) {
			bpp = 32;
			ch = 4;
			if ((bpl / 3) * 3 != bpl)
				return LE_UNSUPPORTEDFORMAT;  // THIS shouldn't happen due to MatchFormat function call...
			bpl += bpl / 3;
			if (bpl > 32767)
				return LE_UNSUPPORTEDSIZE;
		}
	}
	else if (compression != LC_PCT_PACKBITS)
		imageType = OP_DATA_BITS_RECT;
	if (bpp != 1)
		bpl |= 0x8000;

	Write16be(io, imageType);
	if (imageType == OP_DATA_DIRECTBITS_RECT || imageType == OP_DATA_DIRECTBITS_RGN)
		Write32be(io, 0x000000ff);  // according to spec (see A-14 Opcodes in pictures)
	WriteFbe(io, "2 2222 224 44 2222 444", static_cast<lil_u16>(bpl), ZERO_16, ZERO_16, h, w,
		ZERO_16, packType, ZERO_32, dpiX, dpiY,
		pixType, bpp, ch, static_cast<lil_u16>(outFormat.bitsPerChannel),
		ZERO_32, ZERO_32, ZERO_32);
	bpl &= 0x7fff;

	// V2 Src rect, Dest rect, arith transfer mode
	WriteFbe(io, "2222 2222 2", ZERO_16, ZERO_16, h, w, ZERO_16, ZERO_16, h, w, ARITH_SRCCOPY);

	//------------------------
	// 4) SAVE IMAGE DATA
	std::auto_ptr<LilCodecPackbits> codec;
	const LilIoDesc *pIo = io;
	LilBuffer buf;
	if (!buf.Realloc(&image->memFuncs, image->data.height * 4 * 2))
		return LE_NOTENOUGHMEMORY;
	if (compression == LC_PCT_PACKBITS) {
		codec = std::auto_ptr<LilCodecPackbits>(new LilCodecPackbits(io, image, &outFormat, buf.GetPtr<lil_u32>(), false));
		pIo = codec->GetIoPtr();
	}

	err = WriteImageData(pIo, image, &outFormat, true);
	if (err == LE_NONE)
		Write16be(io, OP_ENDPIC);

	return err;
}

} // namespace lil

#endif // LIL_SUPPORTS_PCT
