#include "LilPlatform.h"

#ifdef LIL_SUPPORTS_TIF

#include "LilConvert.h"
#include "LilMem.h"
#include "LilBitManip.h"
#include "LilIo.h"

namespace {
	// These are little endian consts
	const lil_u16 TIFF_ID_LE = 0x4949;  // 0x4550 for mdi tiff?
	const lil_u16 TIFF_ID_BE = 0x4d4d;  // 0x5030 for mdi tiff?
	
	const lil_u16 TIFF_ID_CLASSIC = 0x2a;
	const lil_u16 TIFF_ID_BIG = 0x2b;
}

namespace lil {

//-----------------------------------------------------------------------------------------------------------
bool IsExtensionTif(const tchar *fileName)
{
	if (CheckExtension(fileName, _T(".TIF")))
		return true;
	return false;
}

//-----------------------------------------------------------------------------------------------------------
bool IsDetectableTif(const LilIoDesc *io)
{
	io->seekFunc(io->handle, 0, LS_SET);

	lil_u16 h1 = Read16le(io);
	if (h1 == TIFF_ID_LE || h1 == 0x4550) {
		lil_u16 h2 = Read16le(io);
		if (h2 == TIFF_ID_CLASSIC)
			return true;
	}
	if (h1 == TIFF_ID_BE || h1 == 0x5045) {
		lil_u16 h2 = Read16be(io);
		if (h2 == TIFF_ID_CLASSIC)
			return true;
	}

	io->seekFunc(io->handle, 0, LS_SET);
	return false;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType LoadTif(const LilIoDesc *io, LilImageDesc *image, LilLoadType request, const LilFormatDesc *imageFormat)
{
	LIL_UNUSED_PARAM(io);
	LIL_UNUSED_PARAM(image);
	LIL_UNUSED_PARAM(request);
	LIL_UNUSED_PARAM(imageFormat);
	return LE_UNSUPPORTEDFILEFORMAT;
}

//-----------------------------------------------------------------------------------------------------------
LilErrorType SaveTif(const LilIoDesc *io, const LilImageDesc *image, const LilFormatDesc *imageFormat)
{
	LIL_UNUSED_PARAM(io);
	LIL_UNUSED_PARAM(image);
	LIL_UNUSED_PARAM(imageFormat);
	return LE_UNSUPPORTEDFILEFORMAT;
}

}  // namespace lil

#endif // LIL_SUPPORTS_TIF
