#ifndef LILMEM_H_INCLUDED
#define LILMEM_H_INCLUDED

#include <new>
#include "Lil.h"
#include "LilIo.h"

namespace lil {

//-----------------------------------------------------------------------------------------------------------
// Global options
extern LilConfigDesc globalConfig;

//-----------------------------------------------------------------------------------------------------------
// Allocate memory using a single, universal pool that was predetermined
/*
LilErrorType MemPoolCreate(LilMemDesc *memFuncs, lil_u32 size);
LilErrorType MemPoolFree(LilMemDesc *memFuncs);
void *MemAlloc(const LilMemDesc *memFuncs, lil_u32 size);
lil_u32 MemGetSize(const LilMemDesc *memFuncs);
template <class T> T *MemNew(const LilMemDesc *memFuncs) { T *pOut = reinterpret_cast<T *>(MemAlloc(memFuncs, sizeof(T))); new (pOut) T(); return pOut; }
template <class T, class U> T *MemNew(const LilMemDesc *memFuncs, U u) { T *pOut = reinterpret_cast<T *>(MemAlloc(memFuncs, sizeof(T))); new (pOut) T(u); return pOut; }
template <class T, class U, class V> T *MemNew(const LilMemDesc *memFuncs, U u, V v) { T *pOut = reinterpret_cast<T *>(MemAlloc(memFuncs, sizeof(T))); new (pOut) T(u, v); return pOut; }
template <class T, class U, class V, class W> T *MemNew(const LilMemDesc *memFuncs, U u, V v, W w) { T *pOut = reinterpret_cast<T *>(MemAlloc(memFuncs, sizeof(T))); new (pOut) T(u, v, w); return pOut; }
template <class T> void MemDelete(T *pObj) { }
*/
template <class T> T *MemNew(const LilMemDesc *) { return new T; }
template <class T, class U> T *MemNew(const LilMemDesc *, U u) { return new T(u); }
template <class T, class U, class V> T *MemNew(const LilMemDesc *, U u, V v) { return new T(u, v); }
template <class T, class U, class V, class W> T *MemNew(const LilMemDesc *, U u, V v, W w) { return new T(u, v, w); }
template <class T> void MemDelete(T *pObj) { delete pObj; }

template <class T> struct RaiiMemDelete { T *pObj; RaiiMemDelete(T *pObj) : pObj(pObj) { } ~RaiiMemDelete() { if (pObj) MemDelete(pObj); } };

// Allocate an image
bool ImageAlloc(LilImageDesc *image, const LilFormatDesc *imageFormat, LilErrorType *pErr);

// These functions do calculations typical for image memory handling.
lil_u32 RoundUpToNearest(lil_u32 input, lil_u32 multiple);
lil_u32 RoundDownToNearest(lil_u32 input, lil_u32 multiple);
lil_u32 CalcRemainder(lil_u32 input, lil_u32 multiple);
lil_u32 CalcBpp(const LilFormatDesc *format);
lil_u32 CalcMinBytesNeeded(const LilFormatDesc *format, lil_u32 numPixels);
lil_u32 CalcMinBitsNeeded(const LilFormatDesc *format, lil_u32 numPixels);
lil_u32 CalcBpl(const LilFormatDesc *format, lil_u32 width);
lil_u32 CalcBitsPerLine(const LilFormatDesc *format, lil_u32 width);
lil_u64 CalcByteSize(const LilFormatDesc *format, lil_u32 width, lil_u32 height);

//-----------------------------------------------------------------------------------------------------------
class LilBuffer {
public:
	LilBuffer();
	~LilBuffer();

	// Returns false if it fails
	// @@ This should just be ALLOC
	bool Realloc(const LilMemDesc *pMemFuncs, size_t size);
	size_t GetSize() const { return mSize; }
	template <class T> T *GetPtr() { return reinterpret_cast<T *>(mHandle); }
	const LilIoDesc *GetIoDescPtr() { return &mIoMem; }

private:
	lil_mem_handle mHandle;
	size_t mSize;
	LilMemReallocFunc ReallocFunc;
	LilIoDesc mIoMem;
};

}  // namespace lil

#endif // LILMEM_H_INCLUDED