// LilPlatform.h - To compensate for different build environments.
// While "Lil.h" handles many platform issues that will need to be
// addressed internally and externally together, some issues are only
// important internally.

#ifndef LILPLATFORM_H_INCLUDED
#define LILPLATFORM_H_INCLUDED

#include "Lil.h"

#ifdef LIL_COMPILER_VS
    #define _T(x) x
#endif

//-----------------------------------------------------------------------------------------------------------
// BIT ORDER
#define LIL_MSB_TO_LSB
//#define LIL_LSB_TO_MSB

//-----------------------------------------------------------------------------------------------------------
// SUPPORTED FILE TYPES
#define LIL_SUPPORTS_BMP
#define LIL_SUPPORTS_ICO
#define LIL_SUPPORTS_CUR
//#define LIL_SUPPORTS_TIF
#define LIL_SUPPORTS_JPG
#define LIL_SUPPORTS_PNG
#define LIL_SUPPORTS_PSD
#define LIL_SUPPORTS_TGA
#define LIL_SUPPORTS_GIF
#define LIL_SUPPORTS_PCT
#define LIL_SUPPORTS_PNM

//-----------------------------------------------------------------------------------------------------------
// COMPILER-DEPENDANT MACROS
#ifdef LIL_COMPILER_VS
	#pragma warning(disable: 4127)
    #define LIL_INLINE __forceinline
    #define LIL_BREAK() __debugbreak()
	#ifdef _DEBUG
		#define LIL_ASSERT(x) do { if (!(x)) __debugbreak(); } while (0)
	#else
		#define LIL_ASSERT(x) void(0)
	#endif
#elif defined(LIL_COMPILER_GNU)
    #include <assert.h>
    #define LIL_INLINE inline
    #define LIL_BREAK() assert(0)
	#ifdef _DEBUG
		#define LIL_ASSERT(x) assert(x)
	#else
	#define LIL_ASSERT(x) while (0)
		#endif
#endif

//-----------------------------------------------------------------------------------------------------------
// HELPER MACROS
#define LIL_ARRAY_SIZE(x) (sizeof(x) / sizeof(x[0]))
#define LIL_UNUSED_PARAM(x) (void)x; struct x
#define LIL_FLOAT_TO_FIXED(x) static_cast<lil_i32>((x) * 65536.0 + .5)
#define LIL_FIXED_TO_FLOAT(x) (static_cast<float>(x) / 65536.0)
#define LIL_INT_TO_FIXED(x) ((x) << 16)
#define LIL_FIXED_TO_INT(x) ((x) >> 16)

//-----------------------------------------------------------------------------------------------------------
// HELPER FUNCTIONS
namespace lil {
template <class T> LIL_INLINE T GetMax(T a, T b) { return a > b ? a : b; }
template <class T> LIL_INLINE T GetMax(T a, T b, T c) { return GetMax(a, GetMax(b, c)); }
template <class T> LIL_INLINE T GetMin(T a, T b) { return a < b ? a : b; }
template <class T> LIL_INLINE T GetMin(T a, T b, T c) { return GetMin(a, GetMin(b, c)); }
template <class T> LIL_INLINE void Swap(T &a, T &b) { T c = a; a = b; b = c; }
}  // namespace lil

#endif // LILPLATFORM_H_INCLUDED
