#ifndef LILUTILFORMAT_H_INCLUDED
#define LILUTILFORMAT_H_INCLUDED

#include "LilPlatform.h"

namespace lil {

class LilBuffer;

//-----------------------------------------------------------------------------------------------------------
// SPECIAL STRUCTS FOR WORKING WITH FORMATS

// Saving #1 - Used (per file format) to determine valid formats to save an image in
struct LilFormatRange {
	LilDataFormatType dataFormat;
	// One, or the other, depending on dataFormat
	lil_u8 bitCounts[32][LIL_MAX_CHANNELS];
	lil_u8 bitArrSize;
	lil_u8 numChannels;
	lil_u8 blockLineByteAlign;
	lil_u8 planarFlags;
	LilChannelOrderType lco;
	lil_u8 flippedFlags;
};

// Saving #2 - Used (per file format) to determine valid formats to save an image in
struct LilSimpleBitsRange {
	lil_u8 totalBits;   // However arrange, total bits should be this
	lil_u8 numChannels;
	lil_u8 blockLineByteAlign;
	lil_u8 planarFlags;
	LilChannelOrderType lco;
	lil_u8 flippedFlags;
};



//-----------------------------------------------------------------------------------------------------------
// CONSTANTS

// Saving #3 - Used to do implicit conversion to reach valid format to save an image in
struct LilChannelSwap {
	lil_u32 count;
	LilChannelOrderType channels[LIL_MAX_CHANNELS];
};
const LilChannelSwap CHANNEL_SWAPS[15] = {
	{ 0, { } },  // LCO_UNKNOWN
	{ 0, { } },  // LCO_SINGLE
	{ 0, { } },  // LCO_YA
	{ 1, { LCO_BGR } },  // LCO_RGB
	{ 1, { LCO_RGB } },  // LCO_BGR
	{ 3, { LCO_ABGR, LCO_RGBA, LCO_BGRA } },  // LCO_ARGB
	{ 3, { LCO_ARGB, LCO_RGBA, LCO_BGRA } },  // LCO_ABGR
	{ 3, { LCO_ARGB, LCO_ABGR, LCO_BGRA } },  // LCO_RGBA
	{ 3, { LCO_ARGB, LCO_ABGR, LCO_RGBA } },  // LCO_BGRA
	{ 0, { } },  // LCO_CMY
	{ 0, { } },  // LCO_CMYA
	{ 0, { } },  // LCO_CMYK
	{ 0, { } },  // LCO_CMYKA
	{ 0, { } },  // LCO_YCBCR
	{ 0, { } },  // LCO_YCBCRA
};

// Used to detect when we need to do a true colorspace conversion
struct ColorSpacesEntry {
	lil_u8 colorSpace[15];
	lil_u8 count;
};
const ColorSpacesEntry COLOR_SPACES[] = {
	{ { LCO_UNKNOWN }, 1 },
	{ { LCO_SINGLE, LCO_YA }, 2 },
	{ { LCO_RGB, LCO_BGR, LCO_RGBA, LCO_BGRA, LCO_ARGB, LCO_ABGR }, 6 },
	{ { LCO_CMY, LCO_CMYA }, 2 },
	{ { LCO_CMYK, LCO_CMYKA }, 2 },
	{ { LCO_YCBCR, LCO_YCBCRA }, 2 },
};
const lil_u32 COLOR_SPACE_COUNT = LIL_ARRAY_SIZE(COLOR_SPACES);
// Used for conversion. Conversion will be always to one of these forms. Initially, and then
// reordered.
struct ColorSpaceDefault {
	LilChannelOrderType noAlpha;
	LilChannelOrderType yesAlpha;
};
const ColorSpaceDefault COLOR_SPACE_DEFAULTS[] = {
	{ LCO_UNKNOWN, LCO_UNKNOWN },
	{ LCO_SINGLE, LCO_YA },
	{ LCO_RGB, LCO_RGBA },
	{ LCO_CMY, LCO_CMYA },
	{ LCO_CMYK, LCO_CMYKA },
	{ LCO_YCBCR, LCO_YCBCRA },
};

// Used for writing out values to a file format using function "WriteF"
const lil_u32 ZERO_32 = 0;
const lil_u16 ZERO_16 = 0;
const lil_u8 ZERO_8 = 0;

// Special channel numbers
const lil_u8 LIL_CHANNEL_NONE = 0xff;
const lil_u8 LIL_CHANNEL_ERROR = 0xfe;  // e for error
const lil_u8 LIL_CHANNEL_CONVERT = 0xfc;  // c for convert

// Used to indicate possible values in "ranges".
// See structs "LilFormatRange" and "LilSimpleBitsRange".
const lil_u8 FLAG_YES = 1;
const lil_u8 FLAG_NO = 2;
const lil_u8 FLAG_YESNO = FLAG_YES | FLAG_NO;

// Used for IsFormatDifferent function
const lil_u32 IFD_IGNORE_STRIDE = 0x01;
const lil_u32 IFD_IGNORE_PLANAR = 0x02;
const lil_u32 IFD_IGNORE_FLIPPED = 0x04;

// This has to match LilChannelOrderType in lil.h
const lil_u8 LIL_NUM_CHANNELS[15] = {
	LIL_MAX_CHANNELS, // LCO_UNKNOWN
	1, // LCO_SINGLE
	2, // LCO_YA
	3, // LCO_RGB
	3, // LCO_BGR
	4, // LCO_ARGB
	4, // LCO_ABGR
	4, // LCO_RGBA
	4, // LCO_BGRA
	3, // LCO_CMY
	4, // LCO_CMYA
	4, // LCO_CMYK
	5, // LCO_CMYKA
	3, // LCO_YCBCR
	4, // LCO_YCBCRA
};

// Also has to match LilChannelOrderType in lil.h
const lil_u8 LIL_ALPHA_CHANNEL[15] = {
	LIL_CHANNEL_NONE, // LCO_UNKNOWN
	LIL_CHANNEL_NONE, // LCO_SINGLE
	1, // LCO_YA
	LIL_CHANNEL_NONE, // LCO_RGB
	LIL_CHANNEL_NONE, // LCO_BGR
	0, // LCO_ARGB
	0, // LCO_ABGR
	3, // LCO_RGBA
	3, // LCO_BGRA
	LIL_CHANNEL_NONE, // LCO_CMY
	3, // LCO_CMYA
	LIL_CHANNEL_NONE, // LCO_CMYK
	4, // LCO_CMYKA
	LIL_CHANNEL_NONE, // LCO_YCBCR
	3, // LCO_YCBCRA
};



//-----------------------------------------------------------------------------------------------------------
// BASIC FORMAT MANIPULATION AND UTILITES

// Returns false if an error occurred
bool SetError(LilErrorType err, LilErrorType *pErr);

// See Lil.h for details on the parameters.
// Calls IsFormatValid internally as well.
bool SetFormat(LilFormatDesc *pFormat, LilDataFormatType dataFormat, lil_u8 bitsPerChannel,
	lil_u8 numChannels, lil_u8 blockLineByteAlign, bool planar, LilChannelOrderType channelOrder,
	bool flipped, LilErrorType *pErr);
bool SetExtendedInfo(LilExtendedDesc *pExt, lil_u32 numPages, lil_u32 compression,
	LilDpuType dpu, lil_u32 dpuX, lil_u32 dpuY, LilErrorType *pErr);
bool SetImageSize(LilDataDesc *pDataDesc, lil_u32 width, lil_u32 height, LilErrorType *pErr);
// pPalette is output, pImage has palette with original palette, pFormat is result pal format,
// numEntries is size of palette, pTempBuf just a pointer to some LilBuffer (needed for
// automatic memory management in output palette), pErr is any error output.
// Return value is true if succesfull, false if not.
bool SetPalette(LilPaletteDesc *pPalette, const LilImageDesc *pImage, const LilFormatDesc *pFormat,
    lil_u32 numEntries, LilBuffer *pTempBuf, LilErrorType *pErr);

// Makes format as simple as possible if it uses an unnecessarily complex description.
// Returns true if CHANGED.
bool SimplifyFormat(LilFormatDesc *pFormat);

// For instance, you can't have a channel order of LCO_SINGLE combined with > 1 numChannels.
bool IsFormatValid(const LilFormatDesc *pFormat, LilErrorType *pErr);

// True if they're different.
// If one of the pointers is NULL, than no conversion can be done between formats, so they
// are NOT considered different.
// "flags" are defined above at IFD_IGNORE_*** and can be combined with bitwise-or operation
bool IsFormatDifferent(const LilFormatDesc *pSrcFormat, const LilFormatDesc *pDestFormat, lil_u32 flags = 0);
bool IsColorSpaceDifferent(LilChannelOrderType lco1, LilChannelOrderType lco2);
LilChannelOrderType GetDefaultColorspace(LilChannelOrderType lco);

bool HasAlpha(const LilFormatDesc *pFormat);
lil_u32 GetAlphaBits(const LilFormatDesc *pFormat);
lil_u32 GetChannelBits(const LilFormatDesc *pFormat, lil_u8 channel);

// Sets basic format given the bpp the way that is native to windows.  Meaning
// 16 bits is considered 1555.
bool SetFormatFromBppWin(LilFormatDesc *format, lil_u8 bpp, bool planar, LilErrorType *pErr);
bool SetFormatFromBppTga(LilFormatDesc *format, lil_u8 attrBits, lil_u8 bpp, bool alignBottom, LilErrorType *pErr);

// Returns false if it's invalid or non-existent.
bool CheckValue(lil_i32 value, const lil_i32 *valids, lil_u8 numValids, const lil_i32 *invalids, lil_u8 numInvalids,
	LilErrorType invalidErr, LilErrorType noneErr, LilErrorType *outErr);

// Put supported formats in order of priority.  Definite matches will automatically work, but
// potential matches will default to the first potential match that was matched.
// pRanges (which support direct definitions of SimpleBits) is checked first.  Then pBitRanges
// is checked.
bool MatchFormat(LilFormatDesc *pFormat, const LilFormatRange *pRanges, lil_u32 rangeCount,
	const LilSimpleBitsRange *pBitRanges, lil_u32 bitRangeCount, LilErrorType *pErr);

}  // namespace lil

#endif // LILUTILFORMAT_H_INCLUDED
